! This example shows how the forEachSubTable() function
! is used.
! This subroutine will fill the subtable with dummy data.
subroutine fill(tab)
  use dal

  type(TableT), intent(in) :: tab

  type(ColumnT) :: xCol, yCol, tCol
  real(kind=SINGLE), dimension(:), pointer :: x,y,t

  write(*,*) from( tab ), count( tab )

  xCol = column(tab,"x",MODIFY)
  yCol = column(tab,"y",MODIFY)
  tCol = column(tab,"t",MODIFY)
  x => real32Data(xCol)
  x = 1.23
  write(*,*) x
  y => real32Data(yCol)
  y = 2.34
  write(*,*) y
  t => real32Data(tCol)
  t = 3.45
  write(*,*) t

end subroutine fill

! This subroutine will write the contents of the subtable to standard output.
subroutine check(tab)
  use dal

  type(TableT), intent(in) :: tab

  type(ColumnT) :: xCol, yCol, tCol
  real(kind=single), dimension(:), pointer :: x,y,t

  write(*,*) from( tab ), count( tab )

  xCol = column(tab,"x",READ)
  yCol = column(tab,"y",READ)
  tCol = column(tab,"t",READ)
  x => real32Data(xCol)
  y => real32Data(yCol)
  t => real32Data(tCol)
  write(*,*) "DATA:", x, y, t

end subroutine check

program example_foreachsubtable

  use dal

  implicit none

  ! This part of the program will apply reportX to a table. 
  type(DataSetT) :: set
  type(TableT) :: tab
  type(ColumnT) :: xCol, yCol, tCol
  real(kind=SINGLE), dimension(:), pointer :: x,y,t

  interface
    subroutine fill( subtab )
      use dal
      implicit none
      type(TableT), intent(in) :: subtab
    end subroutine fill
    subroutine check( subtab )
      use dal
      implicit none
      type(TableT), intent(in) :: subtab
    end subroutine check
  end interface

  set = dataSet("test.dat",CREATE)
  tab = addTable(set,"events",10)
  xCol = addColumn(tab,"x",real32,"mm")
  yCol = addColumn(tab,"y",real32,"mm")
  tCol = addColumn(tab,"t",real32,"s")

  call forEachSubTable(tab,fill)
  call forEachSubTable(tab,check)

  call release(set)

end program example_foreachsubtable
