! This example shows how the label, relabel, name and rename interfaces are used.
subroutine displayLabelled( l )
  use dal

  implicit none

  type(LabelledT), intent(in) :: l

  write(*,*) "the object with name ", name( l ), " has label: ", label(l)

end subroutine displayLabelled

subroutine display( set )
  use dal

  implicit none

  type(DataSetT) set
  type(ArrayT) arr
  type(TableT) tab
  type(ColumnT) col
  type(AttributeT) att

  att = attribute( set, 0 )
  write(*,*) name(att), label( att )
  call displayLabelled( labelled( att ) )

  arr = array( set, 0, READ )
  write(*,*) name(arr), label( arr )
  call displayLabelled( labelled( arr ) )

  att = attribute( arr, 0 )
  write(*,*) name(att), label( att )
  call displayLabelled( labelled( att ) )

  tab = table( set, 1 )
  write(*,*) name(tab), label( tab )
  call displayLabelled( labelled( tab ) )

  att = attribute( tab, 0 )
  write(*,*) name(att), label( att )
  call displayLabelled( labelled( att ) )

  col = column( tab, 0, READ )
  write(*,*) name(col), label( col )
  call displayLabelled( labelled( col ) )

  att = attribute( col, 0 )
  write(*,*) name(att), label( att )
  call displayLabelled( labelled( att ) )

end subroutine display

program example_labelled

  use dal

  implicit none

  type(DataSetT) set
  type(ArrayT) arr
  type(TableT) tab
  type(ColumnT) col
  ! type(AttributeT) att
  ! integer(kind=int32), dimension(:,:,:), pointer :: a
  integer, dimension(3), parameter :: s = (/ 3,4,2 /)

  ! create a set
  set = dataSet("test.dat",CREATE)
  call setAttribute(set,"att1","value1","a dataset attribute comment")
  arr = addArray(set, "array", INTEGER32, comment="an array comment", dimensions=s )
  call setAttribute(arr,"att2","value2","an array attribute comment")
  tab = addTable(set, "table", 10, comment="a table comment" )
  call setAttribute(tab,"att3","value3","a table attribute comment")
  col = addColumn(tab,"int8",INTEGER8,comment="a column comment")
  call setAttribute(col,"TLMAX","value4","a column attribute comment")
   
  call display( set )
  call relabel( tab, "a new table comment" )
  call rename( col, "newcolnm" )
  call display( set )

  call release( set )

end program example_labelled
