! This example shows how to use the real64Array2Data interface.
! In the example a dataset is created (opened) containing
! a table with 2 2-dimensional arrays.
!
! The second array has the same data type as the first; this
! is ensured by using the arrayDataType() function to determine
! the data type of the first array.
!
! The columns are then  initialised, on a row-by-row
! basis (i.e. accessing the column's data cell-by-cell),
! before the dataset is released (closed).
program example_arrayarray2data

  use dal
  use errorhandling

  implicit none

  type(DataSetT) set
  type(TableT) tab
  type(ArrayT) arr1, arr2
  real(kind=DOUBLE), dimension(:,:), pointer :: a1, a2
  integer, dimension(2), parameter :: s = (/ 3,4 /)
  integer :: i,j,n

  ! create a set
  set = dataSet("test.dat",CREATE)
  arr1 = addArray( set, "array1", REAL64, s, "km", "array comment" )
  arr2 = addArray( set, "array2", REAL64, s, "km", "array comment" )

  ! fill with unique numbers

  n = 0
  a1 => real64Array2Data(arr1)
  a2 => real64Array2Data(arr2)
  do j=0,3
    do i=0,2 
      a1(i,j) = n
      a2(i,j) = a1(i,j)
      n = n + 1
    end do
  end do

  call release(arr1)
  call release(arr2)
  call release(set)

end program example_arrayarray2data
