! This example shows how to set the data in
! a variable length column.
program example_setdata

  use dal

  implicit none

  integer, parameter :: nRows = 10
  integer, parameter :: maxCellSize = 100
  integer, dimension(0) :: zeroSize
  integer(kind=INT32) :: i
  type(DataSetT) :: set
  type(TableT) :: tab
  type(ColumnT) :: i8col1, i8col2, i16col1, i16col2, i32col1, i32col2
  type(ColumnT) :: r32col1, r32col2, r64col1, r64col2
  type(ColumnT) :: scol1, scol2, bcol1, bcol2

  logical(kind=bool), dimension(maxCellSize) :: b
  integer(kind=INT8), dimension(maxCellSize) :: i8
  integer(kind=INT16), dimension(maxCellSize) :: i16
  integer(kind=INT32), dimension(maxCellSize) :: i32
  real(kind=SINGLE), dimension(maxCellSize) :: r32
  real(kind=DOUBLE), dimension(maxCellSize) :: r64
  character(len=maxCellSize) :: s

  real(kind=SINGLE), dimension(:), pointer :: r32Data

  s = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
  do i = 1, maxCellSize
    i8(i) = i
    i16(i) = i
    i32(i) = i
    r32(i) = i
    r64(i) = i
    b(i) = ((i / 2).eq.0 )
  end do

  set = dataSet("test.dat",Create)
  tab = addTable(set,"someTable",nRows)

  bcol1 = addColumn(tab,"bcol1",Boolean, &
	dimensions=zeroSize,comment="bool data")

  bcol2 = addColumn(tab,"bcol2",Boolean, &
	dimensions=zeroSize,comment="bool data")

  i8col1 = addColumn(tab,"i8col1",Integer8,units="m", &
	dimensions=zeroSize,comment="int8 data")

  i8col2 = addColumn(tab,"i8col2",Integer8,units="m", &
	dimensions=zeroSize,comment="int8 data")

  i16col1 = addColumn(tab,"i16col1",Integer16,units="m", &
	dimensions=zeroSize,comment="int16 data")

  i16col2 = addColumn(tab,"i16col2",Integer16,units="m", &
	dimensions=zeroSize,comment="int16 data")

  i32col1 = addColumn(tab,"i32col1",Integer32,units="m", &
	dimensions=zeroSize,comment="int32 data")

  i32col2 = addColumn(tab,"i32col2",Integer32,units="m", &
	dimensions=zeroSize,comment="int32 data")

  r32col1 = addColumn(tab,"r32col1",Real32,units="m", &
	dimensions=zeroSize,comment="real32 data")

  r32col2 = addColumn(tab,"r32col2",Real32,units="m", &
	dimensions=zeroSize,comment="real32 data")

  r64col1 = addColumn(tab,"r64col1",Real64,units="m", &
	dimensions=zeroSize,comment="real64 data")

  r64col2 = addColumn(tab,"r64col2",Real64,units="m", &
	dimensions=zeroSize,comment="real64 data")

  scol1 = addColumn(tab,"scol1",String,units="m", &
	dimensions=zeroSize,comment="string data")

  scol2 = addColumn(tab,"scol2",String, &
	dimensions=zeroSize,comment="string data")

  do i=0,nRows - 1
  	call setData( bcol1, i, b( 1 : i + 1 ))
  	call setData( bcol2, i, b( 1 : nRows - i ))
  	call setData( i8col1, i, i8( 1 : i + 1 ))
  	call setData( i8col2, i, i8( 1 : nRows - i ))
  	call setData( i16col1, i, i16( 1 : i + 1 ))
  	call setData( i16col2, i, i16( 1 : nRows - i ))
  	call setData( i32col1, i, i32( 1 : i + 1 ))
  	call setData( i32col2, i, i32( 1 : nRows - i ))
  	call setData( r32col1, i, r32( 1 : i + 1 ))
  	call setData( r32col2, i, r32( 1 : nRows - i ))
  	call setData( r64col1, i, r64( 1 : i + 1 ))
  	call setData( r64col2, i, r64( 1 : nRows - i ))
  	call setData( scol1, i, s( 1 : i + 1 ))
  	call setData( scol2, i, s( 1 : nRows - i ))
  end do

  call release( set )

  set = dataSet("test.dat",Modify)
  tab = table(set,"someTable")

  bcol1 = column(tab,"bcol1",Read)
  bcol2 = column(tab,"bcol2",Read)
  i8col1 = column(tab,"i8col1",Read)
  i8col2 = column(tab,"i8col2",Read)
  i16col1 = column(tab,"i16col1",Read)
  i16col2 = column(tab,"i16col2",Read)
  i32col1 = column(tab,"i32col1",Read)
  i32col2 = column(tab,"i32col2",Read)
  r32col1 = column(tab,"r32col1",Read)
  r32col2 = column(tab,"r32col2",Read)
  r64col1 = column(tab,"r64col1",Read)
  r64col2 = column(tab,"r64col2",Read)
  scol1 = column(tab,"scol1",Read)
  scol2 = column(tab,"scol2",Read)

  do i = 0, nRows - 1
    write(*,*) boolData( bcol1, i )
    write(*,*) boolData( bcol2, i )
    write(*,*) int8Data( i8col1, i )
    write(*,*) int8Data( i8col2, i )
    write(*,*) int16Data( i16col1, i )
    write(*,*) int16Data( i16col2, i )
    write(*,*) int32Data( i32col1, i )
    write(*,*) int32Data( i32col2, i )
    write(*,*) real32Data( r32col1, i )
    write(*,*) real32Data( r32col2, i )
    write(*,*) real64Data( r64col1, i )
    write(*,*) real64Data( r64col2, i )
    write(*,*) stringCell( scol1, i )
    write(*,*) stringCell( scol2, i )
  end do

  call release( set )

end program example_setdata

