
/*************************************************************** 
* 
* Model.h XMM Science Analysis System (SAS) 
* 
* Class representing the DAL Memory Model
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: Model.h,v 1.2 2004/03/05 14:43:01 eojero Exp $ 
* 
*****************************************************************/ 


#ifndef MODEL_H
#define MODEL_H


#include "stringutil.h"


class Model
{
	public :
		enum Type { High, HighLow, Low, Unspecified };
		Model( Type type = Unspecified ) : _type( type ) { PROC ENDPROC }
		Model( const char * model ) {
			PROC
			type( model );
			ENDPROC
		}
		Type type() const { PROC RETURN _type; }
		void	type( Type toThisType ) { PROC _type = toThisType; ENDPROC }
		void	type( const char * toThisType ) {
			PROC
			if( !toThisType ) _type = Unspecified;
			else if( uppercase( toThisType ) == "HIGH" ) _type = High;
			else if( uppercase( toThisType ) == "HIGHLOW" ) _type = HighLow;
			else if( uppercase( toThisType ) == "LOW" ) _type = Low;
			else {
				cerr << "Unrecognised memory model: " << toThisType <<  endl;
			  	exit( 1 );
			} // if
			ENDPROC
		}
		std::string typeAsString() {
			PROC
			std::string	typeStr;
			switch( _type ) {
			case High :
				typeStr = "HIGH";
				break;
			case HighLow :
				typeStr = "HIGHLOW";
				break;
			case Low :
				typeStr = "LOW";
				break;
			case Unspecified :
				typeStr = "UNSPECIFIED";
				break;
			default :
				cerr << "Unrecognised model type: " << int( _type ) <<
				endl;
				exit( 1 );
			} // switch
			RETURN typeStr;

		}

	private :
		Type	_type;
};




#endif


