using namespace std;
/* dal_implementation.cc      & XMM Science Analysis System (SAS) 
* 
* Implementation of the dal.f90 API
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: dal_implementation.cc,v 1.3 2005/06/06 09:00:10 jhoar Exp $ 
* 
*****************************************************************/ 
#include <cstdio>	// C

#include <string>	// C++
#include <iostream>
#include <map>

#include <f90cxx.h>	// SAS
#include <stringutil.h>
#include <Tracer.h>
#include <Dal.h>
//#include <DalExt.h>
#include <ExtDal.h>
#include <Handle.h>
#include <errstr.h>
#include <dal_utilities.h>

//#include "DataClerk.h"


// ------------------------ STATICS ----------------------------

static void
unsignedWarning(
	DataComponent*	dc,
	Labelled*	l,
	Dal::DataType	asType )
{
	PROC
	errstr << "The data in the object with qualified name " << l -> qualifiedName();
	errstr << " contains unsigned " << toString( dc -> dataType() );
	errstr << " data, but is being accessed as " << toString( asType ) << " data.";
	errstr << endl << "An " << toString( asType ) << " pointer to the data is being returned";
	errstr << endl << "NOTE: The DAL may treat this as an error in the future.";
	errstr << Sas::warning( "unsignedAccessedAsSigned" );
	ENDPROC
}


// **********************************************************
// We use a librarian to do the book keeping of data objects.
// **********************************************************
//DataClerk dataClerk;


F90MODPROCDEF(Logical,dal,clobber)()
{
	PROC
	RETURN dataSetServer -> clobber();
}


// Conversions from derived type to base type

// Subclasses of Attributable

F90MODPROCDEF(Handle<Attributable>,dal,datasetattributable)(
	const Handle<DataSet> &		attrH )
{
	PROC
	RETURN attrH.addressOfObject();
}

F90MODPROCDEF(Handle<Attributable>,dal,blockattributable)(
	const Handle<Block> &		attrH )
{
	PROC
	RETURN attrH.addressOfObject();
}


F90MODPROCDEF(Handle<Attributable>,dal,arrayattributable)(
	const Handle<Array> &		attrH )
{
	PROC
	RETURN attrH.addressOfObject();
}

F90MODPROCDEF(Handle<Attributable>,dal,tableattributable)(
	const Handle<Table> &		attrH )
{
	PROC
	RETURN attrH.addressOfObject();
}

F90MODPROCDEF(Handle<Attributable>,dal,columnattributable)(
	const Handle<Column> &		attrH )
{
	PROC
	RETURN attrH.addressOfObject();
}


// Subclasses of Block

F90MODPROCDEF(Handle<Block>,dal,arrayblock)(
	const Handle<Array> &		blockH )
{
	PROC
	RETURN blockH.addressOfObject();
}

F90MODPROCDEF(Handle<Block>,dal,tableblock)(
	const Handle<Table> &		blockH )
{
	PROC
	RETURN blockH.addressOfObject();
}


// Subclasses of Labelled

F90MODPROCDEF(Handle<Labelled>,dal,attributablelabelled)(
	const Handle<Attributable> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,datasetlabelled)(
	const Handle<DataSet> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,blocklabelled)(
	const Handle<Block> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,tablelabelled)(
	const Handle<Table> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,arraylabelled)(
	const Handle<Array> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,columnlabelled)(
	const Handle<Column> &		labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

F90MODPROCDEF(Handle<Labelled>,dal,attributelabelled)(
	const Handle<Attribute> &	labelledH )
{
	PROC
	RETURN labelledH.addressOfObject();
}

// Subclasses of Nullable

F90MODPROCDEF(Handle<Nullable>,dal,arraynullable)(
	const Handle<Array> &		nullableH )
{
	PROC
	RETURN nullableH.addressOfObject();
}


F90MODPROCDEF(Handle<Nullable>,dal,columnnullable)(
	const Handle<Column> &		nullableH )
{
	PROC
	RETURN nullableH.addressOfObject();
}


F90MODPROCDEF(Handle<Nullable>,dal,datacomponentnullable)(
	const Handle<DataComponent> &		nullableH )
{
	PROC
	RETURN nullableH.addressOfObject();
}


// Subclasses of DataComponent

F90MODPROCDEF(Handle<DataComponent>,dal,columndatacomponent)(
	const Handle<Column> &		datacomponentH )
{
	PROC
	RETURN datacomponentH.addressOfObject();
}

F90MODPROCDEF(Handle<DataComponent>,dal,arraydatacomponent)(
	const Handle<Array> &		datacomponentH )
{
	PROC
	RETURN datacomponentH.addressOfObject();
}


// Operations on a dataset

F90MODPROCDEF(Handle<DataSet>,dal,dataset)(
	const char *				name,
	const DataSet::AccessMode &		openMode,
	const DataSetServer::MemoryModel *	memoryModel, // optional
	int					nameLength )
{
	PROC

	string setName = trim(string(name,nameLength));

	DataSet * dataSet = memoryModel ?
				dataSetServer -> open( setName, openMode, *memoryModel )
				:
				dataSetServer -> open( setName, openMode );

	// ************************************************************
	// Construct a handle with dataSet,
	// and return it. If the memory model was specified then use it
	// otherwise use the default memory model as given in the C++
	// API.
	// ************************************************************
	RETURN dataSet;
}


F90MODPROCDEF(Handle<DataSet>,dal,clone)(
	const char *				from,
	const char *				to,
	const DataSet::AccessMode &		openMode,
	const DataSetServer::MemoryModel *	memoryModel, // optional
	int					fromLength,
	int					toLength )
{
	PROC

	string fromName = trim( string( from, fromLength ));
	string toName = trim( string( to, toLength ));

	DataSet * dataSet = memoryModel ?
				dataSetServer -> clone( fromName, toName, openMode, *memoryModel )
				:
				dataSetServer -> clone( fromName, toName, openMode );

	// ************************************************************
	// Construct a handle with dataSet,
	// and return it. If the memory model was specified then use it
	// otherwise use the default memory model as given in the C++
	// API.
	// ************************************************************
	RETURN dataSet;
}


F90MODPROCDEF(void,dal,releasedataset)(
	const Handle<DataSet> &	dataSetH )
{
	PROC

	// Make sure all data objects are deleted.
	DataSet* set = dataSetH.addressOfObject();

	//dataClerk.release( set );
	set -> release();
	dataSetServer -> close( set );

	ENDPROC
}


F90MODPROCDEF(void,dal,keepdataset)(
	const char *		name,
	int			nameLength )
{
	PROC
	dataSetServer->keep( trim(string(name,nameLength)) );
	ENDPROC
}


F90MODPROCDEF(void,dal,cleanupclient)(
	const char *		clientname,
	int			clientnameLength )
{
	PROC
	dataSetServer -> cleanupClient( trim( string( clientname, clientnameLength )));
	ENDPROC
}


F90MODPROCDEF(void,dal,discarddataset)(
	const char *		name,
	int			nameLength )
{
	PROC
	dataSetServer->discard( trim(string(name,nameLength)) );
	ENDPROC
}


F90MODPROCDEF(void,dal,deletedataset)(
	const char *		setName,
	int			setNameLength )
{
	PROC
	dataSetServer->remove( trim(string(setName,setNameLength)) );
	ENDPROC
}


F90MODPROCDEF(int,dal,setexists)(
	const char *		name,
	int			nameLength )
{
	PROC
	RETURN dataSetServer->exists( trim(string(name,nameLength)) );
}


F90MODPROCDEF(void,dal,copydataset)(
	const char *		orgName,
	const char *		destName,
	int			orgNameLength,
	int			destNameLength )
{
	PROC
	dataSetServer->copy(
		trim(string(orgName,orgNameLength)),
		trim(string(destName,destNameLength)) );
	ENDPROC
}

F90MODPROCDEF(void,dal,releasetable)(
	const Handle<Table> & tableH)
{
	PROC

	// Delete the data objects associated with the columns.

	//dataClerk.release( tableH.addressOfObject() );
	tableH.addressOfObject() -> release();
	

	ENDPROC

}


F90MODPROCDEF(void,dal,releasecolumn)(
	const Handle<Column> & columnH )
{
	PROC

	// Delete the data object associated with the column.

	//dataClerk.release( columnH.addressOfObject() );
	columnH.addressOfObject() -> release();

	ENDPROC

}


F90MODPROCDEF(void,dal,releasearray)(
	const Handle<Array> & arrayH )
{
	PROC

	// Delete the data object associated with the array

	//dataClerk.release( arrayH.addressOfObject() );
	arrayH.addressOfObject() -> release();

	ENDPROC

}


F90MODPROCDEF(void,dal,releaseblock)(
	const Handle<Block> & blockH)
{
	PROC

	// Delete the data object associated with the array

	//dataClerk.release( blockH.addressOfObject() );
	blockH.addressOfObject() -> release();

	ENDPROC
}



F90MODPROCDEF(DataSet::AccessMode,dal,mode)(
	const Handle<DataSet> &	dataSetH )
{
	PROC

	RETURN dataSetH.addressOfObject() -> mode();

}


F90MODPROCDEF(Logical,dal,hasblock)(
	const Handle<DataSet> &	dataSetH,
	const char *		name,
	int			nameLength )
{
	PROC

	DataSet * set = dataSetH.addressOfObject();

	RETURN set->hasBlock( trim(string(name,nameLength)) );
}

F90MODPROCDEF(void,dal,copyblock)(
	const Handle<DataSet> &	dataSetH,
	const Handle<Block> &	blockH,
	char *			newName,         // optional
	int			newNameLength )
{
	PROC

	DataSet * set = dataSetH.addressOfObject();
	Block * block = blockH.addressOfObject();

	if (newName)
		set->add(
			block,
			newName?trim(string(newName,newNameLength)):string("") );
	else
		set->add(block);

	ENDPROC
}

F90MODPROCDEF(Handle<Table>,dal,addtable)(
	const Handle<DataSet> &		dataSetH,
	const char *			name,
	const int & 			numberOfRows,
	const char *			comment,	// optional
	const int *			position,	// optional
	int				nameLength,
	int				commentLength )
{
	PROC

	DataSet *	dataSetP = dataSetH.addressOfObject();
	const Table * dataTable = dataSetP -> addTable(
							trim(string( name, nameLength )),
							numberOfRows,
							comment ? trim(string(comment,commentLength)) : string(""),
							position ? *position : -1);


	RETURN dataTable;

}


F90MODPROCDEF(int,dal,blocknumber)(
	const Handle<DataSet> &	dataSetH,
	const char *				name,
	int					nameLength )
{
	PROC

	RETURN dataSetH.addressOfObject()->blockNumber( trim(string( name, nameLength ) ) );

}


F90MODPROCDEF(Handle<Array>,dal,addarray)(
	const Handle<DataSet> &		dataSetH,
	const char *			name,
	const Data::DataType &		dataType,
        const Dope<unsigned,1>*	        dimensions,
	const char *			units,		// optional
	const char *			comment,	// optional
	const int *			position,	// optional
	int				nameLength,
	int				unitsLength,
	int				commentLength )
{
	PROC

	string			theName = trim(string( name, nameLength ));
	string			theComment = (comment ? trim(string( comment, commentLength )) : string(""));
	string			theUnits = (units ? trim(string( units, unitsLength )) : string(""));
	DataSet *		dataSetP = dataSetH.addressOfObject();
	const unsigned*	        dim = dimensions -> addr;

 	// Revert the Fortran indices.
	unsigned long		vectorSize = dimensions -> dim[ 0 ].extent;
	vector<unsigned long>	theDimensions( vectorSize );

	for (unsigned i = 0; i < vectorSize; ++i)
		theDimensions[i] = dim[vectorSize-1-i];

	const int		thePosition = (position ? *position : -1 );

	const Array * array = dataSetP -> addArray(
						theName,
						dataType,
						theDimensions,
						theUnits,
						theComment,
						thePosition );

	RETURN array;

}


F90MODPROCDEF(int,dal,numberofdimensionsofarray)(
		const Handle<Array> &	arrayH )
{
 	PROC

	RETURN arrayH.addressOfObject() -> dimensions();

}


F90MODPROCDEF(int,dal,numberofdimensionsofcolumn)(
		const Handle<Column> & columnH )
{
 	PROC
	
	RETURN columnH.addressOfObject() -> dimensions();

}

typedef Dope<unsigned,1> DopeU_1;
F90MODPROCDEF(DopeU_1,dal,dimensionsofarray)(
 	const Handle<Array> &	arrayH )
{
 	PROC

	Array *array = arrayH.addressOfObject();

	unsigned long dimensions = array -> dimensions();

	static unsigned dim[10];
	for (unsigned i = 0; i < dimensions; ++i)
		dim[i] = array->size()[dimensions-i-1];

	RETURN Dope<unsigned,1>(
		dim,
		&dimensions );
}


F90MODPROCDEF(DopeU_1,dal,dimensionsofcolumn)(
 	const Handle<Column> &	columnH )
{
 	PROC

	Column *column = columnH.addressOfObject();

	unsigned long dimensions = column -> dimensions();

	static unsigned dim[10];
	for (unsigned i = 0; i < dimensions; ++i)
		dim[i] = column->size()[dimensions-i-1];

	RETURN Dope<unsigned,1>(
		dim,
		&dimensions );
}


F90MODPROCDEF(int,dal,numberofelementsofarray)(
 	const Handle<Array> &	arrayH )
{
 	PROC

	Array *arrayP = arrayH.addressOfObject();

	RETURN arrayP -> elements();

}


F90MODPROCDEF(int,dal,numberofelementsofcolumn)(
 	const Handle<Column> &	columnH )
{
 	PROC

	Column *column = columnH.addressOfObject();

	RETURN column -> elements();

}

F90MODPROCDEF(int,dal,numberofelementsofcell)(
	const Handle<Column> & columnH,
	int & row) 
{
  PROC
    Column * column = columnH.addressOfObject();
  CellData * cd = column->cellData(row);
  int result = cd->size();
  delete cd;
  RETURN result;
}



F90MODPROCDEF(int,dal,numberofrowsintable)(
	const Handle<Table> &	ThisH )
{
 	PROC
 	RETURN ThisH.addressOfObject() -> rows();
}


F90MODPROCDEF(int,dal,numberofrowsincolumn)(
	const Handle<Column> &	ThisH )
{
 	PROC
 	RETURN ThisH.addressOfObject() -> rows();
}


F90MODPROCDEF(void,dal,deleteblockwithnumber)(
	const Handle<DataSet> &	dataSetH,
	const int &				BlockNumber )
{
	PROC

	dataSetH.addressOfObject() -> deleteBlock( BlockNumber );

	ENDPROC
}


F90MODPROCDEF(void,dal,deleteblockwithname)(
	const Handle<DataSet> &	dataSetH,
	const char *				name,
	int					nameLength )
{
	PROC

	string		blockname = trim(string( name, nameLength ));

	dataSetH.addressOfObject() -> deleteBlock( blockname );

	ENDPROC

}


F90MODPROCDEF(int,dal,numberofblocks)(
	const Handle<DataSet> &	dataSetH )
{
	PROC
	RETURN dataSetH.addressOfObject() -> blocks();
}


F90MODPROCDEF(Block::Type,dal,blocktype)(
	const Handle<Block> &	blockH )
{
	PROC
	RETURN blockH.addressOfObject() -> type();
}


F90MODPROCDEF(Block::Type,dal,blocktypeofblockwithnumber)(
	const Handle<DataSet> &		dataSetH,
	const int &			blockNumber )
{
	PROC

	DataSet *dataSetP = dataSetH.addressOfObject();

	RETURN dataSetP -> block( blockNumber ) -> type();

}


F90MODPROCDEF(Block::Type,dal,blocktypeofblockwithname)(
	const Handle<DataSet> &		dataSetH,
	const char *			blockName,
	int				blockNameLength )
{
	PROC

	DataSet *dataSetP = dataSetH.addressOfObject();
	string	name = trim(string( blockName, blockNameLength ));

	RETURN dataSetP -> block( name ) -> type();

}


F90MODPROCDEF(Handle<Table>,dal,tablewithnumber)(
		const Handle<DataSet> &		dataSetH,
		const int &			tableNumber,
		Dal::AccessMode *		mode )	// optional
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();

	RETURN ( mode ?
			dataSet -> table( tableNumber, *mode )
			:
			dataSet -> table( tableNumber ) );
}


F90MODPROCDEF(Handle<Table>,dal,tablewithname)(
		const Handle<DataSet> &		dataSetH,
		const char *			name,
		Dal::AccessMode *		mode,	// optional
		int				nameLength )
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();
	const string	cleanName = trim( string( name, nameLength ));

	RETURN ( mode ?
			dataSet -> table( cleanName, *mode )
			:
			dataSet -> table( cleanName ) );

}


F90MODPROCDEF(Handle<Array>,dal,arraywithnumber)(
		const Handle<DataSet> &		dataSetH,
		const int &			arrayNumber,
		Dal::AccessMode *		mode )	// optional
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();

	RETURN ( mode ?
			dataSet -> array( arrayNumber, *mode )
			:
			dataSet -> array( arrayNumber ) );
}


F90MODPROCDEF(Handle<Array>,dal,arraywithname)(
		const Handle<DataSet> &	dataSetH,
		const char *			name,
		Dal::AccessMode *		mode,	// optional
		int				nameLength )
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();
	const string	cleanName = trim( string( name, nameLength ));

	RETURN ( mode ?
			dataSet -> array( cleanName, *mode )
			:
			dataSet -> array( cleanName ) );
}


F90MODPROCDEF(Handle<Block>,dal,blockwithnumber)(
		const Handle<DataSet> &		dataSetH,
		const int &			blockNumber,
		Dal::AccessMode *		mode )	// optional
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();

	RETURN ( mode ?
			dataSet -> block( blockNumber, *mode )
			:
			dataSet -> block( blockNumber ) );
}


F90MODPROCDEF(Handle<Block>,dal,blockwithname)(
		const Handle<DataSet> &		dataSetH,
		const char *			name,
		Dal::AccessMode *		mode,	// optional
		int				nameLength )
{
	PROC

	DataSet*	dataSet = dataSetH.addressOfObject();
	const string	cleanName = trim( string( name, nameLength ));

	RETURN ( mode ?
			dataSet -> block( cleanName, *mode )
			:
			dataSet -> block( cleanName ) );
}


F90MODPROCDEF(Handle<Column>,dal,addcolumn)(
		const Handle<Table> & 		dataTableH,
		const char * 			name,
		const Data::DataType &		dataType,
		const char *			units,      // optional
		const Dope<unsigned,1>*		dimensions, // optional
		const char *			comment,    // optional
		const int *			position,   // optional
		int 				nameLength,
		int				unitsLength,
		int 				commentLength )
{
	PROC
    
	SAS_DEBUG( clog << "addColumn " << string(name,nameLength) << ' ' << int( dataType ) << ' ' << (void*)units << ' ' << (units ? trim(string(units,unitsLength)) : string("") ) << ' ' << (position ? *position : -1 ) << endl; )

	string		theComment = ( comment ? trim(string( comment, commentLength )) : string("") );
	string		theUnits = ( units ? trim(string( units, unitsLength )) : string("") );
	Table *		dataTable = dataTableH.addressOfObject();

	vector<unsigned long> size;

	if (dimensions) {
		unsigned vectorSize = dimensions -> dim[ 0 ].extent;
		for (unsigned i = 0; i < vectorSize; ++i)
		  size.push_back(dimensions->addr[vectorSize-i-1]);
	}
        else
			size = Dal::scalar;		

	RETURN dataTable -> addColumn(
			trim(string( name, nameLength )),
			dataType,
			theComment,
			theUnits,
			size,
			(position ? *position : -1 ) );
}

F90MODPROCDEF(void,dal,copycolumn)(
		const Handle<Table> & 		tableH,
		const Handle<Column> &		columnH,
		const char *			newName,    // optional
		int 				newNameLength )
{
	PROC
    
	SAS_DEBUG( clog << "copyColumn " << tableH.addressOfObject()->name() << 
' ' << columnH.addressOfObject()->name() << ' ' << (newName ? string(newName,newNameLength) : string() ) << endl; )

	Table *		table = tableH.addressOfObject();
	Column *	column = columnH.addressOfObject();

	if (newName)
		table->add(column,trim(string(newName,newNameLength)));
	else
		table->add(column);

	ENDPROC
}


F90MODPROCDEF(void,dal,deletecolumnwithnumber)(
		const Handle<Table> & 	thisH,
		const int &		columnNumber )
{
	PROC

	thisH.addressOfObject() -> deleteColumn( columnNumber );

	ENDPROC

}


F90MODPROCDEF(void,dal,deletecolumnwithname)(
		const Handle<Table> & 	columnH,
		const char *		name,
		int			nameLength )
{
	PROC

	columnH.addressOfObject() -> deleteColumn(
		 trim(string(name,nameLength)) );

	ENDPROC

}


F90MODPROCDEF(Logical,dal,hascolumn)(
	const Handle<Table> &	TableH,
	const char *		name,
	int			nameLength )
{
	PROC

	Table * tab = TableH.addressOfObject();

	RETURN tab->hasColumn( trim(string(name,nameLength)) );
}


F90MODPROCDEF(int,dal,numberofcolumns)(
		const Handle<Table> & 	columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> columns();

}


F90MODPROCDEF(Handle<Column>,dal,columnwithname)(
		const Handle<Table> &	tableH,
		const char *		name,
		Dal::AccessMode *	mode,	// optional
		int			nameLength )
{
	PROC

	Table*		table = tableH.addressOfObject();
	const string	cleanName = trim( string( name, nameLength ));

	RETURN ( mode ?
			table -> column( cleanName, *mode )
			:
			table -> column( cleanName ) );
}



F90MODPROCDEF(unsigned int,dal,columnnumberoftablecolumn)(
		const Handle<Table> &	tableH,
		const char *		name,
		int			nameLength )
{
	PROC
	RETURN tableH.addressOfObject() -> columnNumber(
		trim(string(name,nameLength)) );

}


F90MODPROCDEF(unsigned int,dal,columnnumberofcolumn)(
		const Handle<Column> &	columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> columnNumber();
}


F90MODPROCDEF(Handle<Column>,dal,columnwithnumber)(
		const Handle<Table> &	tableH,
		const int &		columnNumber,
		Dal::AccessMode *	mode )	// optional
{
	PROC

	Table*	table = tableH.addressOfObject();

	RETURN ( mode ?
			table -> column( columnNumber, *mode )
			:
			table -> column( columnNumber ) );
}


#define datatype(obj,objname) \
F90MODPROCDEF(Data::DataType,dal,obj##datatype)( \
		const Handle<objname> &	objectH ) \
{ \
	PROC \
	RETURN objectH.addressOfObject() -> dataType(); \
}


datatype( datacomponent, DataComponent );
datatype( column, Column );
datatype( array, Array );

F90MODPROCDEF(Attribute::DataType,dal,attributedatatype)(
		const Handle<Attribute> &	objectH )
{
	PROC
	const Attribute * attribute = objectH.addressOfObject();
	//RETURN ::fitsAttributeDataType( attribute );
	RETURN attribute -> dataType();
}


#define attributedatatypen(obj,objname) \
F90MODPROCDEF(Attribute::DataType,dal,obj##attributedatatypen)( \
		const Handle<objname> &	objectH, \
		const char *		name, \
		int			nameLength ) \
{ \
	PROC \
	const string	cleanName = trim( string( name, nameLength )); \
	const Attribute * attribute = objectH.addressOfObject() -> attribute( cleanName ); \
	RETURN attribute -> dataType(); \
}
	//RETURN ::fitsAttributeDataType( attribute );

attributedatatypen(array,Array);
attributedatatypen(attributable,Attributable);
attributedatatypen(block,Block);
attributedatatypen(column,Column);
attributedatatypen(dataset,DataSet);
attributedatatypen(table,Table);


#define attributedatatyper(obj,objname) \
F90MODPROCDEF(Attribute::DataType,dal,obj##attributedatatyper)( \
		const Handle<objname> &	objectH, \
		int			number ) \
{ \
	PROC \
	RETURN objectH.addressOfObject() -> attribute( number ) ->  dataType(); \
}

attributedatatyper(array,Array);
attributedatatyper(attributable,Attributable);
attributedatatyper(block,Block);
attributedatatyper(column,Column);
attributedatatyper(dataset,DataSet);
attributedatatyper(table,Table);


//F90MODPROCDEF(Data::DataType,dal,columndatatype)( const Handle<Column> &	ThisH )
//{
//	PROC
//
//	RETURN ThisH.addressOfObject() -> dataType();
//}


//F90MODPROCDEF(void,dal,columnunits)(
//			char *			value,
//			int			valueLength,
//			const Handle<Column> &	columnH )
//{
//	PROC
//
//	Column *column = columnH.addressOfObject();
//
//	sprintf(
//		value,
//		"%-*.*s",
//		valueLength,
//		valueLength,
//		column -> units().c_str() );
//
//	ENDPROC
//
//}

F90MODPROCDEF(void,dal,copyrows)(
		const Handle<Table> &	tableH,
		const int &		from,
		const int &		to,
		const int *		count )
{
	PROC

	Table *table = tableH.addressOfObject();

	table->copyRows( from, to, (count ? *count : 1) );

	ENDPROC
} 

F90MODPROCDEF(void,dal,insertrows)(
		const Handle<Table> &	tableH,
		const int &		position,
		const int *		count )
{
	PROC

	Table *table = tableH.addressOfObject();

	table->insertRows( position, (count ? *count : 1) );

	ENDPROC
} 

F90MODPROCDEF(void,dal,deleterows)(
		const Handle<Table> &	tableH,
		const int &		from,
		const int *		count )
{
	PROC

	Table *table = tableH.addressOfObject();

	table->deleteRows( from, (count ? *count : 1) );

	ENDPROC
}

F90MODPROCDEF(void,dal,foreachcolumn)(
		const Handle<Table> & 	tableH,
		const void (*callThisFunction)( const Handle<Column> & ) )
{

	PROC


	Table *table = tableH.addressOfObject();

	for(
		unsigned int i = 0;
		i < table -> columns();
		i++ )
	{
		Column *column = table -> column( i );
		callThisFunction( Handle<Column>(column) );

	} // for

	ENDPROC

}


F90MODPROCDEF(void,dal,foreachblock)(
	const Handle<DataSet> &	dataSetH,
	const void (*callThisFunction)( const Handle<Block> & ) )
{

	PROC

	DataSet* set = dataSetH.addressOfObject();

	for(
		unsigned int i = 0;
		i < set -> blocks();
		i++ )
	{
		Block * block = set -> block( i );
		callThisFunction( Handle<Block>(block) );

	} // for

	ENDPROC

}



#ifdef OLDSCALING

F90MODPROCDEF(void,dal,setscalingofcolumn)(
		const Handle<Column> &		columnH,
		real64 &			zero,
		real64 &			scale,
		Data::DataType &		toType )
{

	PROC

	Column *column = columnH.addressOfObject();

	column -> scaleData( zero, scale, toType );

	ENDPROC

}


F90MODPROCDEF(void,dal,setscalingofarray)(
		const Handle<Array> &		arrayH,
		real64 &			zero,
		real64 &			scale,
		Data::DataType &		toType )
{

	PROC

	Array *arrayP = arrayH.addressOfObject();

	arrayP -> scaleData( zero, scale, toType );

	ENDPROC

}


F90MODPROCDEF(double,dal,zeroofarray)(
		const Handle<Array> &		arrayH )
{
	PROC

	Array *arrayP = arrayH.addressOfObject();

	RETURN arrayP -> zero();
}


F90MODPROCDEF(double,dal,zeroofcolumn)(
		const Handle<Column> &	columnH )
{
	PROC

	Column *column = columnH.addressOfObject();

	RETURN column -> zero();
}


F90MODPROCDEF(double,dal,scaleofarray)(
		const Handle<Array> &		arrayH )
{
	PROC

	Array *arrayP = arrayH.addressOfObject();

	RETURN arrayP -> scale();
}


F90MODPROCDEF(double,dal,scaleofcolumn)(
		const Handle<Column> &	columnH )
{
	PROC

	Column *column = columnH.addressOfObject();

	RETURN column -> scale();

}


F90MODPROCDEF(Logical,dal,hasscalingofcolumn)(
		const Handle<Column> &	columnH )
{
	PROC

	Column *column = columnH.addressOfObject();

	RETURN column -> scaled();

}


F90MODPROCDEF(Logical,dal,hasscalingofarray)(
		const Handle<Array> &		arrayH )
{
	PROC

	Array *arrayP = arrayH.addressOfObject();

	RETURN arrayP -> scaled();

}


F90MODPROCDEF(void,dal,unsetscalingofarray)(
		const Handle<Array> &		arrayH,
		Data::DataType &		ToType )
{
	PROC

	Array *arrayP = arrayH.addressOfObject();
	arrayP -> unscaleData( ToType );

	ENDPROC

}


F90MODPROCDEF(void,dal,unsetscalingofcolumn)(
		const Handle<Column> &		columnH,
		Data::DataType &		ToType )
{
	PROC

	Column *ColumnP = columnH.addressOfObject();
	ColumnP -> unscaleData( ToType );

	ENDPROC

}

#endif


// NULL VALUES

F90MODPROCDEF(int,dal,intnullvaluearray)(
	const Handle<Array>&	arrayH )
{
	PROC
	Array* arr = arrayH.addressOfObject();
	RETURN arr -> intNullValue();
}


F90MODPROCDEF(int,dal,intnullvaluecolumn)(
	const Handle<Column>&	columnH )
{
	PROC
	Column* col = columnH.addressOfObject();
	RETURN col -> intNullValue();
}


F90MODPROCDEF(int,dal,intnullvaluedatacomponent)(
	const Handle<DataComponent>&	dataComponentH )
{
	PROC
	DataComponent* dc = dataComponentH.addressOfObject();
	RETURN dc -> intNullValue();
}


F90MODPROCDEF(int,dal,intnullvaluenullable)(
	const Handle<Nullable>&	nullableH )
{
	PROC
	Nullable* nullable = nullableH.addressOfObject();
	RETURN nullable -> intNullValue();
}


F90MODPROCDEF(void,dal,setnullvaluearray)(
	const Handle<Array>&	arrayH,
	int&			value )
{
	PROC

	Array* arr = arrayH.addressOfObject();
	arr -> nullValue( value );

	ENDPROC
}


F90MODPROCDEF(void,dal,setnullvaluecolumn)(
	const Handle<Column>&	columnH,
	int&			value )
{
	PROC

	Column* col = columnH.addressOfObject();
	col -> nullValue( value );

	ENDPROC
}


F90MODPROCDEF(void,dal,setnullvaluedatacomponent)(
	const Handle<DataComponent>&	dataComponentH,
	int&			value )
{
	PROC

	DataComponent* dc = dataComponentH.addressOfObject();
	dc -> nullValue( value );

	ENDPROC
}


F90MODPROCDEF(void,dal,setnullvaluenullable)(
	const Handle<Nullable>&	nullableH,
	int&			value )
{
	PROC

	Nullable* nullable = nullableH.addressOfObject();
	nullable -> nullValue( value );

	ENDPROC
}


F90MODPROCDEF(void,dal,settonullarray)(
	const Handle<Array>&	arrayH,
	int&			position )
{
	PROC

	Array* arr = arrayH.addressOfObject();
	ArrayData* arrdat = arr -> data();
	arrdat -> setToNull( position );
	delete arrdat;

	ENDPROC
}


F90MODPROCDEF(void,dal,settonullcell)(
	const Handle<Column>&	columnH,
	int&			row,
	int&			position )
{
	PROC

	Column * col = columnH.addressOfObject();
	ColumnData* coldat = col -> data( row, 1 );
	coldat -> setToNull( position );
	delete coldat;

	ENDPROC
}


F90MODPROCDEF(void,dal,settonullcolumn)(
	const Handle<Column>&	columnH,
	int&			row )
{
	PROC

	Column * col = columnH.addressOfObject();
	ColumnData* coldat = col -> data( col->from() + row, 1 );
	coldat -> setToNull( 0 );
	delete coldat;

	ENDPROC
}


F90MODPROCDEF(Logical,dal,nulldefinedarray)(
	const Handle<Array>&	arrayH )
{
	PROC
	RETURN arrayH.addressOfObject() -> nullDefined();
}


F90MODPROCDEF(Logical,dal,nulldefinedcolumn)(
	const Handle<Column>&	columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> nullDefined();
}


F90MODPROCDEF(Logical,dal,nulldefineddatacomponent)(
	const Handle<DataComponent>&	dataComponentH )
{
	PROC
	RETURN dataComponentH.addressOfObject() -> nullDefined();
}


F90MODPROCDEF(Logical,dal,nulldefinednullable)(
	const Handle<Nullable>&	nullableH )
{
	PROC
	RETURN nullableH.addressOfObject() -> nullDefined();
}


F90MODPROCDEF(Logical,dal,nulltypearray)(
	const Handle<Array>&	arrayH )
{
	PROC
	RETURN arrayH.addressOfObject() -> nullType();
}


F90MODPROCDEF(Logical,dal,nulltypecolumn)(
	const Handle<Column>&	columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> nullType();
}


F90MODPROCDEF(Logical,dal,nulltypedatacomponent)(
	const Handle<DataComponent>&	dataComponentH )
{
	PROC
	RETURN dataComponentH.addressOfObject() -> nullType();
}


F90MODPROCDEF(Logical,dal,nulltypenullable)(
	const Handle<Nullable>&	nullableH )
{
	PROC
	RETURN nullableH.addressOfObject() -> nullType();
}


F90MODPROCDEF(Logical,dal,hasnullsarray)(
	const Handle<Array>&	arrayH )
{
	PROC
	Array* arr = arrayH.addressOfObject();
	ArrayData* arrdat = arr -> data();
	bool hasNulls = arrdat -> hasNulls();
	delete arrdat;
	RETURN hasNulls;
}


F90MODPROCDEF(Logical,dal,hasnullscolumn)(
	const Handle<Column>&	columnH )
{
	PROC
	Column* col = columnH.addressOfObject();
	ColumnData* coldat = col -> data();
	bool hasNulls = coldat -> hasNulls();
	delete coldat;
	RETURN hasNulls;
}


F90MODPROCDEF(Logical,dal,isnullarray)(
	const Handle<Array>&	arrayH,
	int&			position )
{
	PROC
	Array* arr = arrayH.addressOfObject();
	ArrayData* arrdat = arr -> data();
	bool isNull = arrdat -> isNull( position );
	delete arrdat;
	RETURN isNull;
}


F90MODPROCDEF(Logical,dal,isnullcell)(
	const Handle<Column>&	columnH,
	int&			row,
	int&			position )
{
	PROC
	Column* col = columnH.addressOfObject();
	ColumnData* coldat = col -> data( row, 1 );
	bool isNull = coldat -> isNull( position );
	delete coldat;
	RETURN isNull;
}


F90MODPROCDEF(Logical,dal,isnullcolumn)(
	const Handle<Column>&	columnH,
	int&			row )
{
	PROC
	Column* col = columnH.addressOfObject();
	ColumnData* coldat = col -> data( row, 1 );
	bool isNull = coldat -> isNull( 0 );
	delete coldat;
	RETURN isNull;
}


/*
static int
ToLogical( int value )
{
	PROC

	RETURN value? TRUE : FALSE;
}

*/


F90MODPROCDEF(void,dal,setattributes)(
	const Handle<Attributable> &	attributableH,
	const Handle<Attributable> &	originH )
{
	PROC

	Attributable * attributable = attributableH.addressOfObject();
	Attributable * origin = originH.addressOfObject();

	*attributable = *origin;

	ENDPROC

}

F90MODPROCDEF(void,dal,addattributes)(
	const Handle<Attributable> &	attributableH,
	const Handle<Attributable> &	originH )
{
	PROC

	Attributable * attributable = attributableH.addressOfObject();
	Attributable * origin = originH.addressOfObject();

	attributable->addAttributes( origin );

	ENDPROC

}

/*
// Obsolete
#define copyattributes(obj,objname) \
F90MODPROCDEF(void,dal,obj##copyattributes)( \
	const Handle<objname> &	attributableH, \
	const Handle<objname> &	originH ) \
{ \
	PROC \
 \
	errstr << "using obsolete function copyAttributes" << Sas::message(Msg::Core, Msg::Verbose ); \
 \
	Attributable * attributable = attributableH.addressOfObject(); \
	Attributable * origin = originH.addressOfObject(); \
 \
	attributable->addAttributes( origin ); \
 \
	ENDPROC \
 \
}

copyattributes(attributable,Attributable);
copyattributes(dataset,DataSet);
copyattributes(block,Block);
copyattributes(table,Table);
copyattributes(array,Array);
copyattributes(column,Column);
*/


#define setattributableattribute(obj,objname) \
F90MODPROCDEF(void,dal,obj##setattribute)( \
	const Handle<objname> &		attributableH, \
	const Handle<Attribute> &	attributeH ) \
{ \
	PROC \
 \
	Attribute * attributeP = attributeH.addressOfObject(); \
	objname * attributableP = attributableH.addressOfObject(); \
	attributableP -> addAttribute( attributeP ); \
 \
	ENDPROC \
 \
}

setattributableattribute(attributable,Attributable);
setattributableattribute(dataset,DataSet);
setattributableattribute(block,Block);
setattributableattribute(array,Array);
setattributableattribute(table,Table);
setattributableattribute(column,Column);


template<class T>
static void addComment(
	const Handle<T> &		attributableH,
	const char *			comment,
	int				commentLength )
{
	PROC

	attributableH.addressOfObject() -> addComment(
		trim(string(comment,commentLength)) );

	ENDPROC

}


#define addcomment(type,typename) \
F90MODPROCDEF(void,dal,addcommentto##type)( \
	const Handle<typename> &		attributableH, \
	const char *				comment, \
	int					commentLength ) \
{ \
	PROC \
	addComment( attributableH, comment, commentLength ); \
	ENDPROC \
}


addcomment(attributable,Attributable);
addcomment(dataset,DataSet);
addcomment(block,Block);
addcomment(array,Array);
addcomment(table,Table);


template<class T>
static void addHistory(
	const Handle<T> &		attributableH,
	const char *			history,
	int				historyLength )
{
	PROC

	attributableH.addressOfObject() -> addHistory(
		trim(string(history,historyLength)) );

	ENDPROC

}


#define addhistory(type,typename) \
F90MODPROCDEF(void,dal,addhistoryto##type)( \
	const Handle<typename> &		attributableH, \
	const char *				history, \
	int					historyLength ) \
{ \
	PROC \
	addHistory( attributableH, history, historyLength ); \
	ENDPROC \
}


addhistory(attributable,Attributable);
addhistory(dataset,DataSet);
addhistory(block,Block);
addhistory(array,Array);
addhistory(table,Table);


template<class T>
int
numberOfAttributes(
	const Handle<T> &			attributableH )
{
	PROC
	RETURN attributableH.addressOfObject() -> attributes();
}


#define numberofattributes(type,typename) \
F90MODPROCDEF(int,dal,numberofattributesof##type)( \
	const Handle<typename> &		attributableH ) \
{ \
	PROC RETURN numberOfAttributes( attributableH ); \
}


numberofattributes(attrib,Attributable);
numberofattributes(dataset,DataSet);
numberofattributes(block,Block);
numberofattributes(table,Table);
numberofattributes(array,Array);
numberofattributes(column,Column);


template<class T>
static int
hasAttribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	T * attributableP = attributableH.addressOfObject();

	RETURN attributableP -> hasAttribute( trim(string(name,nameLength)) );
}


#define hasattribute(type,typename) \
F90MODPROCDEF(Logical,dal,type##hasattribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC RETURN hasAttribute( attributableH, name, nameLength ); \
}

hasattribute(attributable,Attributable);
hasattribute(dataset,DataSet);
hasattribute(block,Block);
hasattribute(array,Array);
hasattribute(table,Table);
hasattribute(column,Column);

//extern "C" template int hasAttribute<DataSet>( const Handle<DataSet> &, const char *, int );
//int( *dal_MP_datasethasattribute)( const Handle<DataSet> &, const char *, int) = hasAttribute;
//const int& dal_MP_datasethasattribute = int(hasAttribute);


//extern "C"
//{
//F90MODPROCPTRDEF(int,dal,datasethasattribute)(
//	const Handle<DataSet> &			dataSetH,
//	const char *				name,
//	int					nameLength ) = hasAttribute;
//}



#define attributewithname(obj,objname) \
F90MODPROCDEF(Handle<Attribute>,dal,obj##attributewithname)( \
		const Handle<objname> &		attributableH, \
		const char *			name, \
		Dal::AccessMode *		mode, \
		int				nameLength ) \
{ \
	PROC \
 \
 	Attributable* 	attributable = attributableH.addressOfObject(); \
	const string	cleanName = trim( string( name, nameLength )); \
 \
	RETURN ( mode ? \
			attributable -> attribute( cleanName, *mode ) \
			: \
			attributable -> attribute( cleanName ) ); \
}


attributewithname(attributable,Attributable);
attributewithname(dataset,DataSet);
attributewithname(block,Block);
attributewithname(table,Table);
attributewithname(array,Array);
attributewithname(column,Column);


#define attributewithnumber(obj,objname) \
F90MODPROCDEF(Handle<Attribute>,dal,obj##attributewithnumber)( \
		const Handle<objname> &		attributableH, \
		const int &			number, \
		Dal::AccessMode *		mode ) \
{ \
	PROC \
 \
 	Attributable* 	attributable = attributableH.addressOfObject(); \
 \
	RETURN ( mode ? \
			attributable -> attribute( number, *mode ) \
			: \
			attributable -> attribute( number ) ); \
 \
}

attributewithnumber(attributable,Attributable);
attributewithnumber(dataset,DataSet);
attributewithnumber(block,Block);
attributewithnumber(table,Table);
attributewithnumber(array,Array);
attributewithnumber(column,Column);


#define attributevalue(tp,fn) \
F90MODPROCDEF(tp,dal,tp##attribute)( \
	const Handle<Attribute> &		attributeH ) \
{ \
	PROC \
 \
	Attribute * attribute = attributeH.addressOfObject(); \
 \
	RETURN attribute -> fn(); \
}

attributevalue(int8,asInt);
attributevalue(int16,asInt);
attributevalue(int32,asInt);
attributevalue(real32,asReal);
attributevalue(real64,asReal);


F90MODPROCDEF(void,dal,stringattribute)(
	char *					value,
	int					valueLength,
	const Handle<Attribute> &		attributeH )
{
	PROC

	Attribute *	attributeP = attributeH.addressOfObject();
	const string	attributeValue = attributeP -> asString();

	const int len = min( int( attributeValue.size() ), valueLength );
	memcpy( value, attributeValue.c_str(), len );
	memset( value + len, ' ', valueLength - len );
	//sprintf(
	//	value,
	//	"%-*.*s",
	//	valueLength,
	//	valueLength,
	//	attribute -> asString().c_str() );

	ENDPROC
}


F90MODPROCDEF(int32,dal,booleanattribute)(
	const Handle<Attribute> &		attributeH )
{
	PROC

	Attribute * attribute = attributeH.addressOfObject();

	RETURN attribute -> asBool();	// We assume that asBool()
					// returns a value which is
					// consistant with Fortrans
					// logical type.

}


template<class T>
static int8
int8Attribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asInt();
}


#define int8attribute(type,typename) \
F90MODPROCDEF(int8,dal,int8##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN int8Attribute( attributableH, name, nameLength ); \
}

int8attribute(attributable,Attributable);
int8attribute(dataset,DataSet);
int8attribute(block,Block);
int8attribute(table,Table);
int8attribute(array,Array);
int8attribute(column,Column);


template<class T>
static int16
int16Attribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asInt();
}


#define int16attribute(type,typename) \
F90MODPROCDEF(int16,dal,int16##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN int16Attribute( attributableH, name, nameLength ); \
}

int16attribute(attributable,Attributable);
int16attribute(dataset,DataSet);
int16attribute(block,Block);
int16attribute(table,Table);
int16attribute(array,Array);
int16attribute(column,Column);


template<class T>
static int32
int32Attribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asInt();
}


#define int32attribute(type,typename) \
F90MODPROCDEF(int32,dal,int32##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN int32Attribute( attributableH, name, nameLength ); \
}

int32attribute(attributable,Attributable);
int32attribute(dataset,DataSet);
int32attribute(block,Block);
int32attribute(table,Table);
int32attribute(array,Array);
int32attribute(column,Column);


template<class T>
static real32
real32Attribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asReal();
}


#define real32attribute(type,typename) \
F90MODPROCDEF(real32,dal,real32##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN real32Attribute( attributableH, name, nameLength ); \
}

real32attribute(attributable,Attributable);
real32attribute(dataset,DataSet);
real32attribute(block,Block);
real32attribute(table,Table);
real32attribute(array,Array);
real32attribute(column,Column);


template<class T>
static real64
real64Attribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asReal();
}


#define real64attribute(type,typename) \
F90MODPROCDEF(real64,dal,real64##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN real64Attribute( attributableH, name, nameLength ); \
}

real64attribute(attributable,Attributable);
real64attribute(dataset,DataSet);
real64attribute(block,Block);
real64attribute(table,Table);
real64attribute(array,Array);
real64attribute(column,Column);


template<class T>
static void
stringAttribute(
	char *					value,
	int					valueLength,
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string	theAttributeName = trim( string( name, nameLength ));
	T *		attributableP = attributableH.addressOfObject();
	Attribute * 	attributeP = attributableP -> attribute( theAttributeName );
	const string	attributeValue = attributeP -> asString();
	const int 	len = min( int( attributeValue.size() ), valueLength );

	memcpy( value, attributeValue.c_str(), len );
	memset( value + len, ' ', valueLength - len );
	//sprintf(
	//	value,
	//	"%-*.*s",
	//	valueLength,
	//	valueLength,
	//	attribute -> asString().c_str() );

	ENDPROC

}


#define stringattribute(type,typename) \
F90MODPROCDEF(void,dal,string##type##attribute)( \
	char *					value, \
	int					valueLength, \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	stringAttribute( value, valueLength, attributableH, name, nameLength ); \
	ENDPROC \
}

stringattribute(attributable,Attributable);
stringattribute(dataset,DataSet);
stringattribute(block,Block);
stringattribute(table,Table);
stringattribute(array,Array);
stringattribute(column,Column);


template<class T>
static int32
booleanAttribute(
	const Handle<T> &			attributableH,
	const char *				name,
	int					nameLength )
{
	PROC

	const string theAttributeName = trim(string(name,nameLength));

	T * attributableP = attributableH.addressOfObject();

	Attribute * attribute =
			attributableP -> attribute( theAttributeName );

	RETURN attribute -> asBool();	// We assume that asBool()
					// returns a value which is
					// consistant with Fortrans
					// logical type.

}


#define booleanattribute(type,typename) \
F90MODPROCDEF(int32,dal,boolean##type##attribute)( \
	const Handle<typename> &		attributableH, \
	const char *				name, \
	int					nameLength ) \
{ \
	PROC \
	RETURN booleanAttribute( attributableH, name, nameLength ); \
}


booleanattribute(attributable,Attributable);
booleanattribute(dataset,DataSet);
booleanattribute(block,Block);
booleanattribute(table,Table);
booleanattribute(array,Array);
booleanattribute(column,Column);



F90MODPROCDEF(void,dal,unitsofattribute)(
		char *				value,
		int				valueLength,
		const Handle<Attribute> &	attributeH )
{
	PROC

	Attribute *	attributeP = attributeH.addressOfObject();
	const string	attributeUnits = attributeP -> units();
	const int 	len = min( int( attributeUnits.size() ), valueLength );

	memcpy( value, attributeUnits.c_str(), len );
	memset( value + len, ' ', valueLength - len );
	//sprintf(
	//	value,
	//	"%-*.*s",
	//	valueLength,
	//	valueLength,
	//	attribute -> units().c_str() );

	ENDPROC

}


#define attributethingy(obj,objname,thingy,fn) \
F90MODPROCDEF(void,dal,obj##attribute##thingy)( \
		char *				value, \
		int				valueLength, \
		const Handle<objname> &		attributableH, \
		const char *			name, \
		int				nameLength ) \
{ \
	PROC \
 \
	const string	theAttributeName = trim( string( name, nameLength )); \
	objname *	attributableP = attributableH.addressOfObject(); \
	Attribute *	attributeP = attributableP -> attribute( theAttributeName ); \
 \
 	const string	thingy = attributeP -> fn(); \
	const int 	len = min( int( thingy.size() ), valueLength ); \
 \
	memcpy( value, thingy.c_str(), len ); \
	memset( value + len, ' ', valueLength - len ); \
 \
	ENDPROC \
 \
}

attributethingy(attributable,Attributable,units,units);
attributethingy(dataset,DataSet,units,units);
attributethingy(block,Block,units,units);
attributethingy(array,Array,units,units);
attributethingy(column,Column,units,units);
attributethingy(table,Table,units,units);

attributethingy(attributable,Attributable,comment,label);
attributethingy(dataset,DataSet,comment,label);
attributethingy(block,Block,comment,label);
attributethingy(array,Array,comment,label);
attributethingy(column,Column,comment,label);
attributethingy(table,Table,comment,label);


#define setthisattribute(tp) \
F90MODPROCDEF(void,dal,set##tp##attribute)( \
		const Handle<Attribute> &	attributeH, \
		const tp &			value, \
		const char *			units, \
		const char *			comment, \
		int				unitsLength, \
		int				commentLength ) \
{ \
	PROC \
	Attribute * attribute = attributeH.addressOfObject(); \
\
	*attribute -> parent() -> addAttribute( \
					attribute -> name(), \
					comment ? trim(string(comment,commentLength)) : string(""), \
					units   ? trim(string(units,unitsLength)):string("")) = value; \
\
	ENDPROC \
}


setthisattribute(int8);
setthisattribute(int16);
setthisattribute(int32);
setthisattribute(real32);
setthisattribute(real64);


F90MODPROCDEF(void,dal,setstringattribute)(
		const Handle<Attribute> &	attributeH,
		const char *			value,
		const char *			comment,
		int				valueLength,
		int				commentLength )
{
	PROC

	Attribute * attribute = attributeH.addressOfObject();

	*attribute -> parent() -> addAttribute(
					attribute -> name(),
					comment ? trim(string(comment,commentLength)) : string(""))

			= trim(string(value,valueLength));

	ENDPROC
}


F90MODPROCDEF(void,dal,setbooleanattribute)(
		const Handle<Attribute> &	attributeH,
		const int32 &			value,
		const char *			comment,
		int				commentLength )
{
	PROC

	Attribute * attribute = attributeH.addressOfObject();

	*attribute -> parent() -> addAttribute(
					attribute -> name(),
					comment ? trim(string(comment,commentLength)) : string(""))
	  = bool(value);

	ENDPROC
}


template<class T,class V>
static void
setAttribute(
	const Handle<T> &			attributableH,
	const char *				name,
	const V &				value,
	const char *				units,
	const char *				comment,
	int					nameLength,
	int					unitsLength,
	int					commentLength )
{
	PROC
	T * attributableP = attributableH.addressOfObject();
	*attributableP -> addAttribute(
								   trim(string(name,nameLength)),
								   comment ? trim(string(comment,commentLength)) : string(""),
								   units   ? trim(string(units,unitsLength)):string(""))
	  = value;
	ENDPROC
}


#define setattribute(tp,obj,objname) \
F90MODPROCDEF(void,dal,set##tp##obj##attribute)( \
		const Handle<objname> &		attributableH, \
		const char *			name, \
		const tp &			value, \
		const char *			units, \
		const char *			comment, \
		int				nameLength, \
		int				unitsLength, \
		int				commentLength ) \
{ \
	PROC \
	setAttribute( \
		attributableH, \
		name, \
		value, \
		units, \
		comment, \
		nameLength, \
		unitsLength, \
		commentLength ); \
	ENDPROC \
}


setattribute(int8,attributable,Attributable);
setattribute(int8,dataset,DataSet);
setattribute(int8,block,Block);
setattribute(int8,array,Array);
setattribute(int8,column,Column);
setattribute(int8,table,Table);

setattribute(int16,attributable,Attributable);
setattribute(int16,dataset,DataSet);
setattribute(int16,block,Block);
setattribute(int16,array,Array);
setattribute(int16,column,Column);
setattribute(int16,table,Table);

setattribute(int32,attributable,Attributable);
setattribute(int32,dataset,DataSet);
setattribute(int32,block,Block);
setattribute(int32,array,Array);
setattribute(int32,column,Column);
setattribute(int32,table,Table);

setattribute(real32,attributable,Attributable);
setattribute(real32,dataset,DataSet);
setattribute(real32,block,Block);
setattribute(real32,array,Array);
setattribute(real32,column,Column);
setattribute(real32,table,Table);

setattribute(real64,attributable,Attributable);
setattribute(real64,dataset,DataSet);
setattribute(real64,block,Block);
setattribute(real64,array,Array);
setattribute(real64,column,Column);
setattribute(real64,table,Table);


template<class T>
static void
setStringAttribute(
		const Handle<T> &	attributableH,
		const char *		name,
		const char *		value,
		const char *		comment,
		int			nameLength,
		int			valueLength,
		int			commentLength )
{
	PROC

	T * attributableP = attributableH.addressOfObject();

	*attributableP -> addAttribute(
			trim(string(name,nameLength)),
			comment ? trim(string(comment,commentLength)) : string(""))
		= trim(string(value,valueLength));

	ENDPROC
}


#define setstringattribute(obj,objname) \
F90MODPROCDEF(void,dal,setstring##obj##attribute)( \
		const Handle<objname> &	attributableH, \
		const char *		name, \
		const char *		value, \
		const char *		comment, \
		int			nameLength, \
		int			valueLength, \
		int			commentLength ) \
{ \
	PROC \
	setStringAttribute( \
		attributableH, \
		name, \
		value, \
		comment, \
		nameLength, \
		valueLength, \
		commentLength ); \
	ENDPROC \
}


setstringattribute(attributable,Attributable);
setstringattribute(dataset,DataSet);
setstringattribute(block,Block);
setstringattribute(array,Array);
setstringattribute(column,Column);
setstringattribute(table,Table);


template<class T>
static void
setBooleanAttribute(
		const Handle<T> &		attributableH,
		const char *			name,
		const int32 &			value,
		const char *			comment,
		int				nameLength,
		int				commentLength )
{
	PROC

	T * attributableP = attributableH.addressOfObject();

	*attributableP -> addAttribute(
			trim(string(name,nameLength)),
			comment ? trim(string(comment,commentLength)) : string(""))
	  = bool(value);
				// We assume that the fortran
				// logical type is consistant
				// with Attribute::Bool
	ENDPROC
}

#define setbooleanattribute(obj,objname) \
F90MODPROCDEF(void,dal,setboolean##obj##attribute)( \
		const Handle<objname> &	attributableH, \
		const char *		name, \
		const int32 &		value, \
		const char *		comment, \
		int			nameLength, \
		int			commentLength ) \
{ \
	PROC \
	setBooleanAttribute( \
		attributableH, \
		name, \
		value, \
		comment, \
		nameLength, \
		commentLength ); \
	ENDPROC \
}


setbooleanattribute(attributable,Attributable);
setbooleanattribute(dataset,DataSet);
setbooleanattribute(block,Block);
setbooleanattribute(array,Array);
setbooleanattribute(column,Column);
setbooleanattribute(table,Table);


#define deleteattributewithnumber(obj,objname) \
F90MODPROCDEF(void,dal,delete##obj##attributewithnumber)( \
		const Handle<objname> & 	attributableH, \
		int				attributeNumber ) \
{ \
	PROC \
	objname * attributableP = attributableH.addressOfObject(); \
	attributableP -> deleteAttribute( attributeNumber ); \
	ENDPROC \
}

deleteattributewithnumber(attrib,Attributable);
deleteattributewithnumber(datset,DataSet);
deleteattributewithnumber(block,Block);
deleteattributewithnumber(array,Array);
deleteattributewithnumber(column,Column);
deleteattributewithnumber(table,Table);


#define deleteattributewithname(obj,objname) \
F90MODPROCDEF(void,dal,delete##obj##attributewithname)( \
		const Handle<objname> & 	attributableH, \
		const char *			name, \
		int				nameLength ) \
{ \
	PROC \
	objname * attributableP = attributableH.addressOfObject(); \
	attributableP -> deleteAttribute( trim(string(name, nameLength ))); \
	ENDPROC \
}


deleteattributewithname(attrib,Attributable);
deleteattributewithname(dataset,DataSet);
deleteattributewithname(block,Block);
deleteattributewithname(array,Array);
deleteattributewithname(column,Column);
deleteattributewithname(table,Table);



F90MODPROCDEF(void,dal,deleteattribute)(
		const Handle<Attribute> &	attributeH )
{
	PROC
	Attribute * attribute = attributeH.addressOfObject();
	attribute -> parent() -> deleteAttribute( attribute -> name() );
	ENDPROC
}


#define objectunits(obj,objname) \
F90MODPROCDEF(void,dal,obj##units)( \
			char *			value, \
			int			valueLength, \
			const Handle<objname> &	objectH ) \
{ \
	PROC \
	objname *	object = objectH.addressOfObject(); \
	const string	units = object -> units(); \
	const int 	len = min( int( units.size() ), valueLength ); \
 \
	memcpy( value, units.c_str(), len ); \
	memset( value + len, ' ', valueLength - len ); \
\
	ENDPROC \
}


objectunits(array,Array);
objectunits(column,Column);
objectunits(datacomponent,DataComponent);


//F90MODPROCDEF(void,dal,arrayunits)(
//			char *			value,
//			int			valueLength,
//			const Handle<Array> &	arrayH )
//{
//	PROC
//
//	Array * array = arrayH.addressOfObject();
//
//	sprintf(
//		value,
//		"%-*.*s",
//		valueLength,
//		valueLength,
//		array -> units().c_str() );
//
//	ENDPROC
//
//}


//F90MODPROCDEF(int,dal,arraydatatype)(
//			const Handle<Array> &	arrayH )
//{
//	PROC
//
//	RETURN arrayH.addressOfObject() -> dataType();
//}


// **********************
// Access Array as vector
// **********************

#define vectorarraydata(tp,dim,name) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,tp##name)( const Handle<Array> & 	arrayH ) \
{ \
	PROC \
 \
	Array* arrayP = arrayH.addressOfObject(); \
	unsigned long		size = arrayP -> elements(); \
 \
	RETURN Dope<tp,dim>( \
			arrayP -> data() -> tp##Data(), \
			&size ); \
 \
}


vectorarraydata(int8,1,arraydata);
//vectorarraydata(int16,1,arraydata);
//vectorarraydata(int32,1,arraydata);
vectorarraydata(real32,1,arraydata);
vectorarraydata(real64,1,arraydata);
vectorarraydata(uint16,1,arraydata);
vectorarraydata(uint32,1,arraydata);

#define uvectorarraydata(tp,dim,name,dt) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,tp##name)( const Handle<Array> & 	arrayH ) \
{ \
	PROC \
 \
	Array* arrayP = arrayH.addressOfObject(); \
	unsigned long		size = arrayP -> elements(); \
 \
 	if( arrayP -> dataType() == Array::U##tp ) \
	{ \
		unsignedWarning( arrayP, arrayP, Array::dt ); \
 \
		RETURN Dope<tp,dim>( \
				(tp*)( arrayP -> data() -> u##tp##Data() ), \
				&size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,dim>( \
				arrayP -> data() -> tp##Data(), \
				&size ); \
	} \
 \
}

uvectorarraydata(int16,1,arraydata,Int16);
uvectorarraydata(int32,1,arraydata,Int32);

#define arraydata(tp,dim,name) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,tp##name)( const Handle<Array> & 	arrayH ) \
{ \
	PROC \
 \
	Array* arrayP = arrayH.addressOfObject(); \
 \
	RETURN Dope<tp,dim>( \
			arrayP -> data() -> tp##Data(), \
			arrayP -> size() ); \
 \
}


arraydata(int8,1,arrayvectordata);
//arraydata(int16,1,arrayvectordata);
//arraydata(int32,1,arrayvectordata);
arraydata(real32,1,arrayvectordata);
arraydata(real64,1,arrayvectordata);
arraydata(uint16,1,arrayvectordata);
arraydata(uint32,1,arrayvectordata);

#define uarraydata(tp,dim,name,dt) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,tp##name)( const Handle<Array> & 	arrayH ) \
{ \
	PROC \
 \
	Array* arrayP = arrayH.addressOfObject(); \
 \
 	if( arrayP -> dataType() == Array::U##tp ) \
	{ \
		unsignedWarning( arrayP, arrayP, Array::dt ); \
 \
		RETURN Dope<tp,dim>( \
				(tp*)( arrayP -> data() -> u##tp##Data() ), \
				arrayP -> size() ); \
	} \
	else \
	{ \
		RETURN Dope<tp,dim>( \
				arrayP -> data() -> tp##Data(), \
				arrayP -> size() ); \
	} \
 \
}

uarraydata(int16,1,arrayvectordata,Int16);
uarraydata(int32,1,arrayvectordata,Int32);

arraydata(int8,2,arrayarray2data);
//arraydata(int16,2,arrayarray2data);
//arraydata(int32,2,arrayarray2data);
arraydata(real32,2,arrayarray2data);
arraydata(real64,2,arrayarray2data);
arraydata(uint16,2,arrayarray2data);
arraydata(uint32,2,arrayarray2data);

uarraydata(int16,2,arrayarray2data,Int16);
uarraydata(int32,2,arrayarray2data,Int32);

arraydata(int8,3,arrayarray3data);
//arraydata(int16,3,arrayarray3data);
//arraydata(int32,3,arrayarray3data);
arraydata(real32,3,arrayarray3data);
arraydata(real64,3,arrayarray3data);
arraydata(uint16,3,arrayarray3data);
arraydata(uint32,3,arrayarray3data);

uarraydata(int16,3,arrayarray3data,Int16);
uarraydata(int32,3,arrayarray3data,Int32);


// *******************************************
// The following functions operate on Columns.
// *******************************************

// ==============
// STRING COLUMNS
// ==============

F90MODPROCDEF(void,dal,setstringcell)(
	const Handle<Column> & 		columnH,
	int &				rowNumber,
	const char *			value,
	int	 			length )
{
	PROC

	Column *column = columnH.addressOfObject();

	//ColumnData *data = dataClerk.data(column);
	ColumnData *data = column -> data( rowNumber, 1 );

        char *s = data->stringData();

 	//int len = min( int(data->size()[0]), length );
 	int len = min( int( column -> elements() ), length );

	//strncpy( s + data->position(rowNumber), value, len);
	strncpy( s, value, len);

	delete data;

	ENDPROC
}
/*
F90MODPROCDEF(void,dal,stringcell)(
	char *				value,
	int				length,
	const Handle<Column> & 		columnH,
	int &				rowNumber )
{
	PROC

	Column *column = columnH.addressOfObject();

	ColumnData *data = dataClerk.data(column);

	const char *s = data->stringData() + data->position(rowNumber);

 	int len = min( int(data->size()[0]), length );

	// Make sure we don't write the trailing zero or beyond
	len = min( len, int(strlen(s)) );

	// Fill with spaces.
	memset( value, ' ', length );

	memcpy( value, s, len );

	ENDPROC
}
*/

/*
F90MODPROCDEF(void,dal,stringcell)(
	char *				value,
	int				length,
	const Handle<Column> & 		columnH,
	int &				rowNumber )
{
	PROC

	Column * column = columnH.addressOfObject();
	char * s;
	int len;

	if( column -> cellType() == Column::Fixed )
	{
		//ColumnData *data = dataClerk.data(column);
		ColumnData *data = column -> data();
		s = data->stringData() + data->position(rowNumber);
 		len = min( int(data->size()[0]), length );
	}
	else
	{
		CellData * data = column -> cellData( rowNumber );
		s = data -> stringData();
 		len = min( int(data->size()), length );
		delete data;
	} // if
	// Make sure we don't write the trailing zero or beyond
	len = min( len, int(strlen(s)) );

	// Fill with spaces.
	memset( value, ' ', length );

	memcpy( value, s, len );

	ENDPROC
}
*/


F90MODPROCDEF(void,dal,stringcell)(
	char *				value,
	int				length,
	const Handle<Column> & 		columnH,
	int &				rowNumber )
{
	PROC

	Column * column = columnH.addressOfObject();
	//char * s;
	//int len;

	// FIXME: here we assume a couple of things about the manner
	// used to write string columns. In particular we assume that
	// the following mechanism, enforced by cfitsio, is at work:
	//  . writing a null terminated string, the null is removed,
	//  and the string is pedded with blanks to the column's
	//  width.
	//  . reading such a string, trailing blanks are removed, and
	//  the string appears to be null terminated. It is not clear
	//  whether the null termination is enforced by cfitsio, or is
	//  just an accident.
	//
	// Based on this assumptions, in order to make the fortran
	// trim function work as expected, we make sure that blanks
	// are used to everwrite every trailing character.

	if( column -> cellType() == Column::Fixed )
	{
		ColumnData *data = column -> data( rowNumber, 1 );
		std::string ss(data->stringData());
		// pad with blanks by hand
		for(unsigned long k = ss.size(); k < column->elements(); ++k)
		  ss += ' ';
		// rejoin the original algorithm
		const char* s = ss.c_str();		
 		const int len = min( int( column -> elements() ), length );
		memcpy( value, s, len );
		memset( value + len, ' ', length - len );
		delete data;
	}
	else
	{
		CellData * data = column -> cellData( rowNumber );
		std::string ss(data -> stringData());
		// pad with blanks by hand
		for(unsigned long k = ss.size(); k < data->size(); ++k)
		  ss += ' ';
		// rejoin the original algorithm		
		const char* s = ss.c_str();
 		const int len = min( int(data->size()), length );
		memcpy( value, s, len );
		memset( value + len, ' ', length - len );
		delete data;
	} // if
	// Make sure we don't write the trailing zero or beyond
	//len = min( len, int(strlen(s)) );

	// Fill with spaces.
	//memset( value, ' ', length );

	//memcpy( value, s, len );

	ENDPROC
}



// =======================
// VARIABLE LENGTH COLUMNS
// =======================

#define setcell(tp,fn) \
F90MODPROCDEF(void,dal,set##fn##cell)( \
	const Handle<Column> & 		columnH, \
	int &				rowNumber, \
	const Dope<tp,1>*		value ) \
{ \
	PROC \
 \
	Column *column = columnH.addressOfObject(); \
	CellData * data = column->cellData( rowNumber ); \
 \
	data -> size( value -> dim[ 0 ].extent ); \
	tp * b = data -> fn##Data(); \
 \
	const unsigned long size = data -> size(); \
 \
	for( unsigned int i = 0; i < size; ++i ) \
	{ \
		b[ i ] = value -> addr[ i ]; \
	} \
 \
	delete data; \
 \
	ENDPROC \
}


setcell(bool8,bool);
setcell(int8,int8);
setcell(int16,int16);
setcell(int32,int32);
setcell(real32,real32);
setcell(real64,real64);


F90MODPROCDEF(void,dal,setstringvariablecell)(
	const Handle<Column> & 		columnH,
	int &				rowNumber,
	const char *			value,
	int	 			length )
{
	PROC

	Column *column = columnH.addressOfObject();
	CellData * data = column->cellData( rowNumber );

	data -> size( length + 1 );
        char *s = data->stringData();

	strncpy( s, value, length );

	s[ length ] = '\0';

	ENDPROC
}


// ===========================================
// SCALAR COLUMNS AND VARIABLE LENGTH COLUMNS.
// ===========================================

#define scalarcolumndataelement(tp,fn) \
typedef Dope<tp,1> Dope##tp##_1; \
F90MODPROCDEF(Dope##tp##_1,dal,fn##columndataelement)( \
		const Handle<Column> & 		columnH, \
		int &				rowNumber ) \
{ \
	PROC \
 \
	Column *column = columnH.addressOfObject(); \
	unsigned long size; \
	tp * data; \
 \
	if( column -> cellType() == Column::Fixed ) \
	{ \
		ColumnData * columnData = column -> data( rowNumber, 1 ); \
		data = columnData -> fn##Data(); \
		size = column -> elements(); \
	} \
	else \
	{ \
		CellData * cellData = column -> cellData( rowNumber ); \
		data = cellData -> fn##Data(); \
		size = cellData -> size(); \
	} \
 \
	RETURN Dope<tp,1>( data, &size ); \
 \
}


scalarcolumndataelement(bool8,bool);
scalarcolumndataelement(int8,int8);
//scalarcolumndataelement(int16,int16);
//scalarcolumndataelement(int32,int32);
scalarcolumndataelement(real32,real32);
scalarcolumndataelement(real64,real64);
scalarcolumndataelement(uint16,uint16);
scalarcolumndataelement(uint32,uint32);

// =========================================
// SIGNED AND UNSIGNED SCALAR COLUMNS.
// NOTE: f90 does not support unsigned data;
// we choose to return unsigned data
// as signed.
// =========================================

#define uscalarcolumndataelement(tp,dt) \
typedef Dope<tp,1> Dope##tp##_1; \
F90MODPROCDEF(Dope##tp##_1,dal,tp##columndataelement)( \
		const Handle<Column> & 		columnH, \
		int &				rowNumber ) \
{ \
	PROC \
 \
	Column *column = columnH.addressOfObject(); \
	unsigned long size; \
	tp * data; \
 \
	if( column -> cellType() == Column::Fixed ) \
	{ \
		ColumnData * columnData = column -> data( rowNumber, 1 ); \
 \
 		if( column -> dataType() == Column::U##tp ) \
		{ \
			unsignedWarning( column, column, Column::dt ); \
 \
			data = (tp*)(columnData -> u##tp##Data()); \
		} \
		else \
		{ \
			data = columnData -> tp##Data(); \
		} \
		size = column -> elements(); \
	} \
	else \
	{ \
		CellData * cellData = column -> cellData( rowNumber ); \
\
 		if( column -> dataType() == Column::U##tp ) \
		{ \
			unsignedWarning( column, column, Column::dt ); \
 \
			data = (tp*)(cellData -> u##tp##Data()); \
		} \
		else \
		{ \
			data = cellData -> tp##Data(); \
		} \
		size = cellData -> size(); \
	} \
 \
	RETURN Dope<tp,1>( data, &size ); \
 \
}

uscalarcolumndataelement(int16,Int16);
uscalarcolumndataelement(int32,Int32);


// ======================
// SIGNED SCALAR COLUMNS.
// ======================

#define scalarcolumndata(tp,fn) \
typedef Dope<tp,1> Dope##tp##_1; \
F90MODPROCDEF(Dope##tp##_1,dal,fn##columndata)( const Handle<Column> & columnH) \
{ \
	PROC \
 \
	Column *column = columnH.addressOfObject(); \
 \
	unsigned long size = column -> count() * column -> elements(); \
 \
	RETURN Dope<tp,1>( \
			column -> data() -> fn##Data(), \
			&size ); \
}

scalarcolumndata(bool8,bool);
scalarcolumndata(int8,int8);
//scalarcolumndata(int16,int16);
//scalarcolumndata(int32,int32);
scalarcolumndata(real32,real32);
scalarcolumndata(real64,real64);
scalarcolumndata(uint16,uint16);
scalarcolumndata(uint32,uint32);

// =========================================
// SIGNED AND UNSIGNED SCALAR COLUMNS.
// NOTE: f90 does not support unsigned data;
// we choose to return unsigned data
// as signed.
// =========================================

#define uscalarcolumndata(tp,dt) \
typedef Dope<tp,1> Dope##tp##_1; \
F90MODPROCDEF(Dope##tp##_1,dal,tp##columndata)( const Handle<Column> & columnH) \
{ \
	PROC \
 \
	Column *column = columnH.addressOfObject(); \
 \
	const unsigned long size = column -> count() * column -> elements(); \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,1>( \
				(tp*)( column -> data() -> u##tp##Data() ), \
				&size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,1>( \
				column -> data() -> tp##Data(), \
				&size ); \
	} \
}

uscalarcolumndata(int16,Int16);
uscalarcolumndata(int32,Int32);


// ===========================
// FIXED LENGTH VECTOR COLUMNS
// ===========================

#define columndataelement(tp,dim,name,fn) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,fn##name)( \
		const Handle<Column> & 		columnH, \
		int &				rowNumber ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
 \
	RETURN Dope<tp,dim>( \
			column -> data( rowNumber, 1 ) -> fn##Data(), \
			column -> size() ); \
 \
}

columndataelement(bool8,1,columnvectordataelement,bool);
columndataelement(int8,1,columnvectordataelement,int8);
//columndataelement(int16,1,columnvectordataelement,int16);
//columndataelement(int32,1,columnvectordataelement,int32);
columndataelement(real32,1,columnvectordataelement,real32);
columndataelement(real64,1,columnvectordataelement,real64);
columndataelement(uint16,1,columnvectordataelement,uint16);
columndataelement(uint32,1,columnvectordataelement,uint32);


#define ucolumndataelement(tp,dim,name,fn,dt) \
typedef Dope<tp,dim> Dope##tp##_##dim ; \
F90MODPROCDEF(Dope##tp##_##dim,dal,fn##name)( \
		const Handle<Column> & 		columnH, \
		int &				rowNumber ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,dim>( \
				(tp*)(column -> data( rowNumber, 1 ) -> u##fn##Data()), \
				column -> size() ); \
	} \
	else \
	{ \
		RETURN Dope<tp,dim>( \
				column -> data( rowNumber, 1 ) -> fn##Data(), \
				column -> size() ); \
	} \
 \
}

ucolumndataelement(int16,1,columnvectordataelement,int16,Int16);
ucolumndataelement(int32,1,columnvectordataelement,int32,Int32);


#define columnvectordata(tp,fn) \
typedef Dope<tp,2> Dope##tp##_2 ; \
F90MODPROCDEF(Dope##tp##_2,dal,fn##columnvectordata)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 2 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
 \
	RETURN Dope<tp,2>( \
			column -> data() -> fn##Data(), \
			size ); \
 \
}

columnvectordata(bool8,bool);
columnvectordata(int8,int8);
//columnvectordata(int16,int16);
//columnvectordata(int32,int32);
columnvectordata(real32,real32);
columnvectordata(real64,real64);
columnvectordata(uint16,uint16);
columnvectordata(uint32,uint32);

#define ucolumnvectordata(tp,dt) \
typedef Dope<tp,2> Dope##tp##_2 ; \
F90MODPROCDEF(Dope##tp##_2,dal,tp##columnvectordata)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 2 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,2>( \
				(tp*)( column -> data() -> u##tp##Data() ), \
				size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,2>( \
				column -> data() -> tp##Data(), \
				size ); \
	} \
 \
}

ucolumnvectordata(int16,Int16);
ucolumnvectordata(int32,Int32);


// =============================
// FIXED LENGTH 2D ARRAY COLUMNS
// =============================
columndataelement(bool8,2,columnarray2dataelement,bool);
columndataelement(int8,2,columnarray2dataelement,int8);
//columndataelement(int16,2,columnarray2dataelement,int16);
//columndataelement(int32,2,columnarray2dataelement,int32);
columndataelement(real32,2,columnarray2dataelement,real32);
columndataelement(real64,2,columnarray2dataelement,real64);
columndataelement(uint16,2,columnarray2dataelement,uint16);
columndataelement(uint32,2,columnarray2dataelement,uint32);

ucolumndataelement(int16,2,columnarray2dataelement,int16,Int16);
ucolumndataelement(int32,2,columnarray2dataelement,int32,Int32);


#define columnarray2data(tp,fn) \
typedef Dope<tp,3> Dope##tp##_3 ; \
F90MODPROCDEF(Dope##tp##_3,dal,fn##columnarray2data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 3 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
 \
	RETURN Dope<tp,3>( \
			column -> data() -> fn##Data(), \
			size ); \
 \
}

columnarray2data(bool8,bool);
columnarray2data(int8,int8);
//columnarray2data(int16,int16);
//columnarray2data(int32,int32);
columnarray2data(real32,real32);
columnarray2data(real64,real64);
columnarray2data(uint16,uint16);
columnarray2data(uint32,uint32);


#define ucolumnarray2data(tp,dt) \
typedef Dope<tp,3> Dope##tp##_3 ; \
F90MODPROCDEF(Dope##tp##_3,dal,tp##columnarray2data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 3 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,3>( \
				(tp*)( column -> data() -> u##tp##Data() ), \
				size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,3>( \
				column -> data() -> tp##Data(), \
				size ); \
	} \
 \
}

ucolumnarray2data(int16,Int16);
ucolumnarray2data(int32,Int32);


// =============================
// FIXED LENGTH 3D ARRAY COLUMNS
// =============================
columndataelement(bool8,3,columnarray3dataelement,bool);
columndataelement(int8,3,columnarray3dataelement,int8);
//columndataelement(int16,3,columnarray3dataelement,int16);
//columndataelement(int32,3,columnarray3dataelement,int32);
columndataelement(real32,3,columnarray3dataelement,real32);
columndataelement(real64,3,columnarray3dataelement,real64);
columndataelement(uint16,3,columnarray3dataelement,uint16);
columndataelement(uint32,3,columnarray3dataelement,uint32);

ucolumndataelement(int16,3,columnarray3dataelement,int16,Int16);
ucolumndataelement(int32,3,columnarray3dataelement,int32,Int32);


#define columnarray3data(tp,fn) \
typedef Dope<tp,4> Dope##tp##_4 ; \
F90MODPROCDEF(Dope##tp##_4,dal,fn##columnarray3data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 4 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
	size[ 3 ] = column -> size()[ 2 ]; \
 \
	RETURN Dope<tp,4>( \
			column -> data() -> fn##Data(), \
			size ); \
}

columnarray3data(bool8,bool);
columnarray3data(int8,int8);
//columnarray3data(int16,int16);
//columnarray3data(int32,int32);
columnarray3data(real32,real32);
columnarray3data(real64,real64);
columnarray3data(uint16,uint16);
columnarray3data(uint32,uint32);

#define ucolumnarray3data(tp,dt) \
typedef Dope<tp,4> Dope##tp##_4 ; \
F90MODPROCDEF(Dope##tp##_4,dal,tp##columnarray3data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 4 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
	size[ 3 ] = column -> size()[ 2 ]; \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,4>( \
				(tp*)( column -> data() -> u##tp##Data() ), \
				size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,4>( \
				column -> data() -> tp##Data(), \
				size ); \
	} \
}

ucolumnarray3data(int16,Int16);
ucolumnarray3data(int32,Int32);


// =============================
// FIXED LENGTH 4D ARRAY COLUMNS
// =============================
columndataelement(bool8,4,columnarray4dataelement,bool);
columndataelement(int8,4,columnarray4dataelement,int8);
//columndataelement(int16,4,columnarray4dataelement,int16);
//columndataelement(int32,4,columnarray4dataelement,int32);
columndataelement(real32,4,columnarray4dataelement,real32);
columndataelement(real64,4,columnarray4dataelement,real64);
columndataelement(uint16,4,columnarray4dataelement,uint16);
columndataelement(uint32,4,columnarray4dataelement,uint32);

ucolumndataelement(int16,4,columnarray4dataelement,int16,Int16);
ucolumndataelement(int32,4,columnarray4dataelement,int32,Int32);


#define columnarray4data(tp,fn) \
typedef Dope<tp,5> Dope##tp##_5 ; \
F90MODPROCDEF(Dope##tp##_5,dal,fn##columnarray4data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 5 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
	size[ 3 ] = column -> size()[ 2 ]; \
	size[ 4 ] = column -> size()[ 3 ]; \
 \
	RETURN Dope<tp,5>( \
			column -> data() -> fn##Data(), \
			size ); \
}

columnarray4data(bool8,bool);
columnarray4data(int8,int8);
//columnarray4data(int16,int16);
//columnarray4data(int32,int32);
columnarray4data(real32,real32);
columnarray4data(real64,real64);
columnarray4data(uint16,uint16);
columnarray4data(uint32,uint32);

#define ucolumnarray4data(tp,dt) \
typedef Dope<tp,5> Dope##tp##_5 ; \
F90MODPROCDEF(Dope##tp##_5,dal,tp##columnarray4data)( const Handle<Column> & columnH ) \
{ \
	PROC \
 \
	Column *		column = columnH.addressOfObject(); \
	vector<unsigned long>	size( 5 ); \
 \
	size[ 0 ] = column -> rows(); \
	size[ 1 ] = column -> size()[ 0 ]; \
	size[ 2 ] = column -> size()[ 1 ]; \
	size[ 3 ] = column -> size()[ 2 ]; \
	size[ 4 ] = column -> size()[ 3 ]; \
 \
 	if( column -> dataType() == Column::U##tp ) \
	{ \
		unsignedWarning( column, column, Column::dt ); \
 \
		RETURN Dope<tp,5>( \
				(tp*)( column -> data() -> u##tp##Data() ), \
				size ); \
	} \
	else \
	{ \
		RETURN Dope<tp,5>( \
				column -> data() -> tp##Data(), \
				size ); \
	} \
}

ucolumnarray4data(int16,Int16);
ucolumnarray4data(int32,Int32);


// This class definition used to be inside the foreachsubtable
// function, which was OK until the optimzation compiler
// was used, whereupon, the linking failed with missing
// symbols. The reason for the problem is unknown
// but interestingly, if the 'void proc' function
// is commented out, the problem goes away (of course
// the foreachsubtable function will fail to work properly
// in this case.
	class Processor : public SubTableProcessor {
	public :
		Processor(
			Table *			table,
			void(*callThisFunction)( const Handle<Table> &))
			
			: SubTableProcessor( table ),
			  _callThisFunction( callThisFunction ) { PROC ENDPROC }
		void proc( SubTable * subTable ) {
		  PROC
		  Table * table = ((Child<Table>&)*subTable).parent();
		  unsigned long	oldFrom = table -> from();
		  unsigned long oldCount = table -> count();
		  table -> seek( subTable -> from(), subTable -> count() );
		  _callThisFunction( table );
		  table -> seek( oldFrom, oldCount );
		  ENDPROC
		}

	private :
		void(*_callThisFunction)( const Handle<Table> &);
	};


F90MODPROCDEF(void,dal,foreachsubtable)(
		const Handle<Table> & 		tableH,
		void (*callThisFunction)( const Handle<Table> & ) )
{
	PROC

	// BUG: the operation should really be on a subtable, that only
        // accesses the columns of interest.
	//Table *table = tableH.addressOfObject();
	//SubTable subTable(table);
	//while (subTable.next())
	//	callThisFunction( subTable );
	//callThisFunction( tableH );

	Processor processor( tableH.addressOfObject(), callThisFunction );
	processor.forEachSubTable();


	//Alternatively, could have used:
	//SubTableProcessor	processor( tableH.addressOfObject() );
	//while( processor.next() ) callThisFunction( processor.subTable() );

	ENDPROC
}


F90MODPROCDEF(Handle<SubTable>,dal,subtable)( const Handle<Table> & tableH )
{
	PROC
	RETURN ( new SubTableProcessor( tableH.addressOfObject() ) ) -> subTable();
}


F90MODPROCDEF(Logical,dal,subtablenext)( const Handle<SubTable>& subTableH )
{
	PROC
	SubTable * subTable = subTableH.addressOfObject();
	Child<SubTableProcessor> & c = (Child<SubTableProcessor>&)*subTable;
	SubTableProcessor * subTableProcessor = c.parent();
	RETURN subTableProcessor -> next();
}


//F90MODPROCDEF(void,dal,subtableseek)( const Handle<SubTable>& subTableH, unsigned from, unsigned count)
//{
//	PROC

//	subTableH.addressOfObject()->seek(from,count);

//	ENDPROC
//}

// This class definition used to be inside the foreachrow
// function, which was OK until the optimzation compiler
// was used, whereupon, the linking failed with missing
// symbols. The reason for the problem is unknown
// but interestingly, if the 'void proc' function
// is commented out, the problem goes away (of course
// the foreachrow function will fail to work properly
// in this case.
	class RowProcessor : public SubTableProcessor {
	public :
		RowProcessor(
			Table *			table,
			void(*callThisFunction)( const Handle<Table> &))

			: SubTableProcessor( table ),
			  _callThisFunction( callThisFunction ) { PROC ENDPROC }
		void proc( SubTable * subTable ) {
		  PROC
		  Table * table = ((Child<Table>&)*subTable).parent();
		  table -> seek( subTable -> from(), subTable -> count() );
		  _callThisFunction( table );
		  ENDPROC
		}

	private :
		void(*_callThisFunction)( const Handle<Table> & );
	};



F90MODPROCDEF(void,dal,foreachrow)(
	const Handle<Table> & 		tableH,
	void (*callThisFunction)( const Handle<Table> & ) )
{
	PROC
	RowProcessor processor( tableH.addressOfObject(), callThisFunction );
	processor.forEachRow();


	// BUG: the columns of interest should be specified.
	//Table* table = tableH.addressOfObject();
	//SubTableProcessor row(table);

	//while (row.next())
	  //callThisFunction( row );

	ENDPROC
}

//	sprintf(
//	 	name,
//	 	"%-*.*s",
//	 	nameLength,
//	 	nameLength,
//		attributableP -> fn().c_str() );

// =================================
// Names, Qualified Names and Labels
// =================================

#define attributablethingy(obj,objname,thingy,fn) \
F90MODPROCDEF(void,dal,thingy##of##obj)( \
	char *					name, \
	int					nameLength, \
	const Handle<objname> &			attributableH ) \
{ \
	PROC \
 \
	objname * attributableP = attributableH.addressOfObject(); \
 \
	const int len = min( int( attributableP -> fn().size() ), nameLength ); \
	memcpy( name, attributableP -> fn().c_str(), len ); \
	memset( name + len, ' ', nameLength - len ); \
	ENDPROC \
 \
}

attributablethingy(attributable,Attributable,name,name);
attributablethingy(dataset,DataSet,name,name);
attributablethingy(block,Block,name,name);
attributablethingy(array,Array,name,name);
attributablethingy(labelled,Labelled,name,name);
attributablethingy(table,Table,name,name);
attributablethingy(column,Column,name,name);

attributablethingy(attributable,Attributable,qualifiedname,qualifiedName);
attributablethingy(dataset,DataSet,qualifiedname,qualifiedName);
attributablethingy(block,Block,qualifiedname,qualifiedName);
attributablethingy(array,Array,qualifiedname,qualifiedName);
attributablethingy(labelled,Labelled,qualifiedname,qualifiedName);
attributablethingy(table,Table,qualifiedname,qualifiedName);
attributablethingy(column,Column,qualifiedname,qualifiedName);

attributablethingy(attributable,Attributable,label,label);
attributablethingy(dataset,DataSet,label,label);
attributablethingy(block,Block,label,label);
attributablethingy(array,Array,label,label);
attributablethingy(labelled,Labelled,label,label);
attributablethingy(table,Table,label,label);
attributablethingy(column,Column,label,label);


#define attributeop(op,fn) \
F90MODPROCDEF(void,dal,op##ofattribute)( \
	char *					name, \
	int					nameLength, \
	const Handle<Attribute> &		attributeH ) \
{ \
	PROC \
 \
	Attribute * attributeP = attributeH.addressOfObject(); \
 \
	const int len = min( int( attributeP -> fn().size() ), nameLength ); \
	memcpy( name, attributeP -> fn().c_str(), len ); \
	memset( name + len, ' ', nameLength - len ); \
 \
	ENDPROC \
 \
}

attributeop( name,name );
attributeop( qualifiedname,qualifiedName );
attributeop( label,label );


#define rethingyattributable(obj,objname,thingy) \
F90MODPROCDEF(void,dal,re##thingy##obj)( \
	const Handle<objname>& 		attributableH, \
	const char *			newValue, \
	int				valueLength ) \
{ \
	PROC \
	attributableH.addressOfObject() -> re##thingy( trim(string(newValue,valueLength))); \
	ENDPROC \
}

rethingyattributable(attributable,Attributable,label);
rethingyattributable(dataset,DataSet,label);
rethingyattributable(block,Block,label);
rethingyattributable(array,Array,label);
rethingyattributable(labelled,Labelled,label);
rethingyattributable(table,Table,label);
rethingyattributable(column,Column,label);

rethingyattributable(attributable,Attributable,name);
rethingyattributable(dataset,DataSet,name);
rethingyattributable(block,Block,name);
rethingyattributable(array,Array,name);
rethingyattributable(labelled,Labelled,name);
rethingyattributable(table,Table,name);
rethingyattributable(column,Column,name);


F90MODPROCDEF(void,dal,relabelattribute)(
	const Handle<Attribute>& 	attributeH,
	const char *			newLabel,
	int				length )
{
	PROC
	attributeH.addressOfObject() -> relabel( trim(string(newLabel,length)));
	ENDPROC
}


F90MODPROCDEF(void,dal,renameattribute)(
	const Handle<Attribute>& 	attributeH,
	const char *			newName,
	int				length )
{
	PROC
	attributeH.addressOfObject() -> rename( trim(string(newName,length)));
	ENDPROC
}


#define parent(ret,objname,fn) \
F90MODPROCDEF(Handle<ret>,dal,parent##fn)( \
	const Handle<objname> &	object ) \
{ \
	PROC \
	RETURN object.addressOfObject() -> parent(); \
}

parent(DataSet,Table,datasetoftable);
parent(DataSet,Array,datasetofarray);
parent(DataSet,Block,datasetofblock);
parent(Table,Column,table);
parent(Attributable,Attribute,attributable);


#define setattributableunits(obj,objname) \
F90MODPROCDEF(void,dal,set##obj##attributeunits)( \
	const Handle<objname>& 		attributableH, \
	const char *			attributeName, \
	const char *			units, \
	int				attributeNameLength, \
	int				unitsLength ) \
{ \
	PROC \
	const string	theAttributeName = trim( string( attributeName, attributeNameLength )); \
	const string	theUnits = trim( string( units, unitsLength )); \
	Attribute * attribute = attributableH.addressOfObject() -> attribute( theAttributeName ); \
	attribute -> units( theUnits ); \
	ENDPROC \
}

setattributableunits(array,Array);
setattributableunits(attributable,Attributable);
setattributableunits(block,Block);
setattributableunits(column,Column);
setattributableunits(dataset,DataSet);
setattributableunits(table,Table);


#define setunits(obj,objname) \
F90MODPROCDEF(void,dal,set##obj##units)( \
	const Handle<objname>& 		objectH, \
	const char *			units, \
	int				unitsLength ) \
{ \
	PROC \
	const string	theUnits = trim( string( units, unitsLength )); \
	objectH.addressOfObject() -> units( theUnits ); \
	ENDPROC \
}

setunits(array,Array);
setunits(attribute,Attribute);
setunits(column,Column);


F90MODPROCDEF(int,dal,counttable)( const Handle<Table> & tableH )
{
	PROC
	RETURN tableH.addressOfObject() -> count();
}


F90MODPROCDEF(int,dal,countcolumn)( const Handle<Column> & columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> count();
}


F90MODPROCDEF(int,dal,fromtable)( const Handle<Table> & tableH )
{
	PROC
	RETURN tableH.addressOfObject() -> from();
}


F90MODPROCDEF(int,dal,fromcolumn)( const Handle<Column> & columnH )
{
	PROC
	RETURN columnH.addressOfObject() -> from();
}


F90MODPROCDEF(void,dal,seektable)( \
	const Handle<Table>& 		tableH, \
	const int & 			from, \
	const int & 			count )
{
	PROC

	if (count < 0) {
		errstr << "Negative count in seek: " << count;
        	errstr << Sas::error( "negativeSeekCount" );
	}
	tableH.addressOfObject() -> seek( from, count );
	ENDPROC
}


F90MODPROCDEF(void,dal,seekcolumn)( \
	const Handle<Column>& 		columnH, \
	const int & 			from, \
	const int & 			count )
{
	PROC
        if (count < 0) {
                errstr << "Negative count in seek " << count;
                errstr << Sas::error( "negativeSeekCount" );
        }
	columnH.addressOfObject() -> seek( from, count );
	ENDPROC
}


F90MODPROCDEF(Handle<TableSeeker>,dal,newtableseeker)(
	Handle<Table>&		tableH )
{
	PROC
	RETURN new TableSeeker( tableH.addressOfObject() );
}


F90MODPROCDEF(void,dal,releasetableseeker)(
	Handle<TableSeeker>&		tableSeekerH )
{
	PROC
	delete tableSeekerH.addressOfObject();
	ENDPROC
}


F90MODPROCDEF(int,dal,tableseekernext)(
	Handle<TableSeeker>&		tableSeekerH )
{
	PROC
	TableSeeker* tableSeeker = tableSeekerH.addressOfObject();
	RETURN tableSeeker -> next();
}


F90MODPROCDEF(int,dal,fromtableseeker)( const Handle<TableSeeker>& tableSeekerH )
{
	PROC
	TableSeeker* tableSeeker = tableSeekerH.addressOfObject();
	RETURN tableSeeker -> from();
}


F90MODPROCDEF(int,dal,counttableseeker)( const Handle<TableSeeker>& tableSeekerH )
{
	PROC
	TableSeeker* tableSeeker = tableSeekerH.addressOfObject();
	RETURN tableSeeker -> count();
}


F90MODPROCDEF(void,dal,resettableseeker)( const Handle<TableSeeker>& tableSeekerH )
{
	PROC
	TableSeeker* tableSeeker = tableSeekerH.addressOfObject();
	tableSeeker -> reset();
	ENDPROC
}


F90MODPROCDEF(void,dal,setdataxxx)( \
        const Handle<Column> & columnTo , \
        const Handle<Column> & columnFrom)
{
	PROC
        setData( columnTo.addressOfObject(), columnFrom.addressOfObject() ) ; 
        ENDPROC
}





// client
F90MODPROCDEF(void,dal,setclient)( \
				   const char * name,
				   int nameLength )
{
  PROC
  std::string theName = trim(std::string(name), nameLength);
  dataSetServer->client(theName);
  ENDPROC
}

F90MODPROCDEF(void,dal,client)( \
				char * name,
				int nameLength )
{
  PROC
  const std::string theName = dataSetServer->client();
  const long len = min(static_cast<int>(theName.size()), nameLength);
  memcpy(name , theName.c_str(), len );
  memset(name + len, ' ', nameLength - len );
  ENDPROC
}
				   
