/*************************************************************** 
* 
* HiLowArray.h               & XMM Science Analysis System (SAS) 
* 
* Class representing block arrays.
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowArray.h,v 1.2 2004/03/05 14:43:02 eojero Exp $
* 
*****************************************************************/ 
#ifndef HILOW_ARRAY_H
#define HILOW_ARRAY_H


#include <Dal.h>

#include "DalArray.h"
#include "HiLowDataStorage.h"
#include "DalSeekable.h"


class HiLowArray : public DalArray, public DalSeekable< longVec >
{
	public :

		// Constructors
		HiLowArray(
			const std::string &			name,
			DataType 			dataType,
			const longVec &			size,
			const std::string &			units,
			const std::string &			label,
			DalDataSet *			parent );

		HiLowArray(
			const Array & 	original,
			DalDataSet * 	parent,
			std::string 		newName = "");

		virtual ~HiLowArray();

		HiLowArray& operator=( const Array& );
 
		unsigned long elements() const;
		unsigned int dataSize() const;
		//ArrayData * data(
		//		const longVec &		from = zero,
		//		const longVec &		to = zero,
		//		AccessMode		accessMode=AsParent) const;

		void scaleData(
				double			zero,
				double			scale,
				DataType		toType );

		void unscaleData(
				DataType		toType );

		void printOn( ostream & os ) const;

		//void release();

	private :

		void	setData( const HiLowArray &	usingDataInThisArray );

		HiLowDataStorage *		_data;

		ArrayData * makeArrayData(
				AccessMode		accessMode ) const;

};



#endif
