/*************************************************************** 
* 
* HiLowColumnData.h            & XMM Science Analysis System (SAS) 
* 
* Class representing the Data in a Column. 
* 
* (c) 1997 ESA - Written by Mark Thomas, CARA
* 
* $Id: HiLowColumnData.h,v 1.2 2004/03/05 14:43:02 eojero Exp $
* 
*****************************************************************/ 
#ifndef HILOW_COLUMN_DATA
#define HILOW_COLUMN_DATA


#include <Dal.h>

#include "HiLowData.h"
#include "HiLowDataStorage.h"
#include "DalColumnData.h"
#include "DalColumn.h"


class HiLowColumnData : public DalColumnData, public HiLowData<DalColumn>
{
	public :

		HiLowColumnData(
			HiLowDataStorage *		data,
			const DalColumn *		parent,
			unsigned long			from,
			unsigned long			count,
			AccessMode			accessMode );

		virtual ~HiLowColumnData();

		int8 * int8Data() const;
		uint16 * uint16Data() const;
		int16 * int16Data() const;
		uint32 * uint32Data() const;
		int32 * int32Data() const;
		real32 * real32Data() const;
		real64 * real64Data() const;
		bool8 * boolData() const;
		char * stringData() const;

    		unsigned long position(
				unsigned long			row) const;

    		unsigned long position(
				unsigned long			row,
				unsigned long		 	pos) const;

    		unsigned long position(
				unsigned long			row,
				const longVec& 			pos) const;

		const longVec& size() const;

		unsigned long elements() const;

		void seek( unsigned long from, unsigned long count );

		void dumpData( char ** values, unsigned long from, unsigned long count ) const;
		void dumpData( bool8 * values, unsigned long from, unsigned long count ) const;
		void dumpData( int8 * values, unsigned long from, unsigned long count ) const;
		void dumpData( uint16 * values, unsigned long from, unsigned long count ) const;
		void dumpData( int16 * values, unsigned long from, unsigned long count ) const;
		void dumpData( uint32 * values, unsigned long from, unsigned long count ) const;
		void dumpData( int32 * values, unsigned long from, unsigned long count ) const;
		void dumpData( real32 * values, unsigned long from, unsigned long count ) const;
		void dumpData( real64 * values, unsigned long from, unsigned long count ) const;
		void loadData( char ** values, unsigned long from, unsigned long count ) const;
		void loadData( bool8 * values, unsigned long from, unsigned long count ) const;
		void loadData( int8 * values, unsigned long from, unsigned long count ) const;
		void loadData( uint16 * values, unsigned long from, unsigned long count ) const;
		void loadData( int16 * values, unsigned long from, unsigned long count ) const;
		void loadData( uint32 * values, unsigned long from, unsigned long count ) const;
		void loadData( int32 * values, unsigned long from, unsigned long count ) const;
		void loadData( real32 * values, unsigned long from, unsigned long count ) const;
		void loadData( real64 * values, unsigned long from, unsigned long count ) const;
		void copyData( unsigned long from, unsigned long to, unsigned long count ) const;

};


#endif


