#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dal.h>


const bool8 true = 1;
const bool8 false = 0;


struct BlockDesc
{
	BlockType	type;
	unsigned int 	blockNumber; } blockDesc[] =
{
  	{ ArrayType, 0 },
  	{ ArrayType, 1 },
 	{ TableType, 2 },
 	{ TableType, 3 },
  	{ ArrayType, 4 }, 
  	{ ArrayType, 5 },
  	{ TableType, 6 },
  	{ TableType, 7 },
  	{ ArrayType, 8 },
  	{ ArrayType, 9 },
  	{ TableType, 10 },
 	{ TableType, 11 },
  	{ ArrayType, 12 },
  	{ ArrayType, 13 }
};


const unsigned int nrBlocks = sizeof( blockDesc ) / sizeof( blockDesc[ 0 ] );

static void
test( DataSet * set, unsigned int offset )
{
	unsigned int i;
	/* All blocks in one pass, before closing set. */
	for( i = 0; i < nrBlocks; ++i )
	{
		Block *		block = blockWithNumber( set, i + offset );
		Name		name;

		switch( blockDesc[ i ].type )
		{
			case TableType :
				sprintf( name, "table%d", i );
				break;

			case ArrayType :
				sprintf( name, "array%d", i );
				break;
			default :
				printf("Invalid  block type\n");
				exit(1);
		} // switch
		// Check block name
		if( strcmp( nameOfBlock( block ), name ) )
		{
			printf( "invalid block name\n" );
			exit( 1 );
		} // if
		// Check block name using position within set
		if( strcmp( nameOfBlock( blockWithNumber( set, blockDesc[ i ].blockNumber  + offset)), name ))
		{
			printf( "wrong block name\n" );
			exit( 1 );
		} // if
		// Check block number: first method
		if( blockNumber( set, nameOfBlock( block )) != blockDesc[ i ].blockNumber  + offset )
		{
			printf( "wrong block number for table\n" );
			exit( 1 );
		} // if
		// Check block number: second method
		//if( block -> parent() -> blockNumber( block -> name() ) != blockDesc[ i ].blockNumber  + offset) errstr << "wrong block number for table" << Sas::error( "internalError" );
	} // for
}


int main()
{

	unsigned int		dimsn = 3;
	unsigned long		dims[ dimsn ];
	const char*		fileName = "test.dat";
	unsigned int		i;
  	DataSet *		set;

	dims[ 0 ] = 2;
	dims[ 1 ] = 3;
	dims[ 2 ] = 4;

	set = dataSet(fileName,Create,UseEnvironment);

	for( i = 0; i < nrBlocks; ++i )
	{
		Block *		block;
		Name		name;
		Label		label;
		Units		units;

		switch( blockDesc[ i ].type )
		{
			case TableType :
			{
				sprintf( name, "table%d", i );
				sprintf( label, "table label%d", i );
				block = tableToBlock( addTable(
						set,
						name,
						100,
						label,
						blockDesc[ i ].blockNumber ));
			}
				break;

			case ArrayType :
			{
				sprintf( name, "array%d", i );
				sprintf( label, "array label%d", i );
				sprintf( units, "array units%d", i );
				block = arrayToBlock( addArray(
						set,
						name,
						Int32,
						dimsn,
						dims,
						label,
						units,
						blockDesc[ i ].blockNumber ));
			}
				break;
			default :
				printf( "Invalid  block type\n" );
				exit( 1 );
		} // switch

		// Check block name
		if( strcmp( nameOfBlock( block ), name ))
		{
			printf( "invalid block name\n" );
			exit( 1 );
		} // if
		// Check block name using position within set
		if( strcmp( nameOfBlock( blockWithNumber( set, blockDesc[ i ].blockNumber ) ),name ))
		{
			printf( "wrong block name\n" );
			exit( 1 );
		} // if
		// Check block number: first method
		if( blockNumber( set, nameOfBlock( block )) != blockDesc[ i ].blockNumber )
		{
			printf( "wrong block number for table\n" );
			exit( 1 );
		} // if
		// Check block number: second method
		//if( block -> parent() -> blockNumber( block -> name() ) != blockDesc[ i ].blockNumber ) errstr << "wrong block number for table" << Sas::error( "internalError" );

	} // for

	// Test with default blocks, while set is still open.
	test( set, 0 );

	releaseDataSet( set );

  	set = dataSet(fileName,Modify,UseEnvironment);
	// Test with default blocks, while set has been reopened
	test( set, 0 );
	releaseDataSet( set );
  	set = dataSet(fileName,Modify,UseEnvironment);
	addTable(
		set,
		"tab0",
		100,
		"",
		0 );
	// Test again, while set is still open
	test( set, 1 );

	releaseDataSet( set );

  	set = dataSet(fileName,Modify,UseEnvironment);
	// Test again, while set has been reopened
	test( set, 1 );

	releaseDataSet( set );

  	set = dataSet(fileName,Modify,UseEnvironment);
	addTable(
		set,
		"tab1",
		100,
		"",
		0 );
	addTable(
		set,
		"tab2",
		100,
		"",
		0 );
	addTable(
		set,
		"tab3",
		100,
		"",
		0 );
	test( set, 4 );
	releaseDataSet( set );

	// Now add the primary array.
  	set = dataSet(fileName,Modify,UseEnvironment);
	addArray(
			set,
			"PRIMARY",
			Int32,
			dimsn,
			dims,
			"",
			"",
			0 );
	test( set, 5 );

	releaseDataSet( set );

  	set = dataSet(fileName,Modify,UseEnvironment);
	test( set, 5 );
	releaseDataSet( set );

	return 0;

}
