using namespace std;
#include<cassert>

#include "Dal.h"

static void
basicChecks( DataSet * clone, Dal::AccessMode mode )
{
	assert( dataSetServer -> exists( "test1.dat"  ));
	assert( clone -> name() == "test1.dat" );
	assert( clone -> mode() == mode );
}


static void
setAttributes( Attributable* set, const string& prepend )
{
  	*set->addAttribute(prepend + "bool","bool comment") = false;
  	*set->addAttribute(prepend + "int","int comment","int unit") = 1;
  	*set->addAttribute(prepend + "real","real comment","real unit") = 2.5;
  	*set->addAttribute(prepend + "string","string comment") = "a set string";

}


static void
checkAttributes( Attributable* set, const string& prepend )
{
	assert( set -> hasAttribute( prepend + "bool" ) );
	assert( set -> hasAttribute( prepend + "int" ) );
	assert( set -> hasAttribute( prepend + "real" ) );
	assert( set -> hasAttribute( prepend + "string" ) );

	assert( set -> attribute( prepend + "bool" ) -> dataType() == Attribute::Bool );
	assert( set -> attribute( prepend + "int" ) -> dataType() == Attribute::Int );
	assert( set -> attribute( prepend + "real" ) -> dataType() == Attribute::Real );
	assert( set -> attribute( prepend + "string" ) -> dataType() == Attribute::DString );

}


static void
doTest1( Dal::AccessMode mode, unsigned long nRows, const vector<unsigned long>& size )
{
	DataSet* set;

	set = dataSetServer -> open( "test.dat", Dal::Create );
	dataSetServer -> close( set );
	DataSet* clone;
	clone = dataSetServer -> clone( "test.dat", "test1.dat", mode );
	basicChecks( clone, mode );
	dataSetServer -> close( clone );

	set = dataSetServer -> open( "test.dat", Dal::Create );
	setAttributes( set, "s" );
	checkAttributes( set, "s" );

	Array* arr = set -> addArray( "PRIMARY", Array::Real64, size, "km", "array comment" );
	ArrayData* ad = arr -> data();
	real64* r64 = ad -> real64Data();

	for( unsigned int i = 0; i < arr -> elements(); ++ i)
	{
		r64[ i ] = i;
	} // for

	delete ad;

	setAttributes( arr, "a" );
	checkAttributes( arr, "a" );

	Table* tab = set -> addTable( "table", nRows, "table comment" );
	setAttributes( tab, "t" );
	checkAttributes( tab, "t" );

	Column* col = tab -> addColumn( "col1", Column::Int32, "1st column", "mm" );
	ColumnData* cd = col -> data();
	int32* i32 = cd -> int32Data();

	for( unsigned int i = 0; i < col -> elements() * col -> rows(); ++i )
	{
		i32[ i ] = i;
	} // for

	delete cd;

	col = tab -> addColumn( "col2", Column::Int32, "1st column", "mm" );
	cd = col -> data();
	i32 = cd -> int32Data();

	for( unsigned int i = 0; i < col -> elements() * col -> rows(); ++i )
	{
		i32[ i ] = i;
	} // for

	delete cd;

	dataSetServer -> close( set );

	clone = dataSetServer -> clone( "test.dat", "test1.dat", mode );
	basicChecks( clone, mode );
	checkAttributes( clone, "s" );
	assert( clone -> blocks() == 2 );
	assert( clone -> block(0) -> type() == Block::ArrayT );
	assert( clone -> block(0) -> name() == "PRIMARY" );
	assert( clone -> block(1) -> type() == Block::TableT );
	assert( clone -> block(1) -> name() == "table" );

	arr = clone -> array( "PRIMARY" );
	checkAttributes( arr, "a" );
	ad = arr -> data();
	r64 = ad -> real64Data();

	for( unsigned int i = 0; i < arr -> elements(); ++ i)
	{
		assert( r64[ i ] == i );
	} // for

	delete ad;

	tab = clone -> table( "table" );
	assert( tab -> rows() == nRows );
	checkAttributes( tab, "t" );

	assert( tab -> columns() == 2 );
	col = tab -> column( "col1" );
	cd = col -> data();
	i32 = cd -> int32Data();

	for( unsigned int i = 0; i < col -> elements() * col -> rows(); ++i )
	{
		assert( i32[ i ] == int32(i) );
	} // for

	col = tab -> column( "col2" );
	cd = col -> data();
	i32 = cd -> int32Data();

	for( unsigned int i = 0; i < col -> elements() * col -> rows(); ++i )
	{
		assert( i32[ i ] == int32(i) );
	} // for

	delete cd;

	dataSetServer -> close( clone );

}


static void
doTest( unsigned long nRows, const vector<unsigned long>& size )
{
	doTest1( Dal::Temp, nRows, size );
	doTest1( Dal::Modify, nRows, size );
	DataSet* set = dataSetServer -> open( "test1.dat", Dal::Read );
	basicChecks( set, Dal::Read );
	checkAttributes( set, "s" );
	dataSetServer -> close( set );
}


int
main()
{
	vector<unsigned long> size;
	size.push_back( 2 );
	size.push_back( 3 );
	size.push_back( 4 );
	doTest( 0, size );
	doTest( 1, size );
	doTest( 1000, size );
}
