using namespace std;
#include<cassert>

#include <Dal.h>

static struct Cols {
  	char *			name;
	Column::DataType	type;
	char * 			units;
	char *			label;
} cols[] =
{
  	{ "bool",	Column::Bool,		"",	"bool data" },
  	{ "string",	Column::DString,	"",	"string data" },
 	{ "int8",	Column::Int8,		"m",	"int8 data" },
	{ "uint16",	Column::Uint16,		"m",	"uint16 data" },
	{ "int16",	Column::Int16,		"m",	"int16 data" },
	{ "uint32",	Column::Uint32,		"m",	"uint32 data" },
  	{ "int32",	Column::Int32,		"m",	"int32 data" },
  	{ "real32",	Column::Real32,		"m",	"real32 data" },
  	{ "real64",	Column::Real64,		"m",	"real64 data" },
};

static const unsigned numberOfColumns = sizeof( cols ) / sizeof( cols[ 0 ] );
static unsigned long nRows;


template<class T>
static void
fillData( T * data, unsigned long size )
{
     for( unsigned long i = 0; i < size; ++i )
     {
     	data[ i ] = i + size;
     } // for
}


static void
fill( Table * table )
{
	for( unsigned int i = 0; i < table -> columns(); ++i )
	{
		Column * col = table -> column( i );
  		assert(col->elements() == 1);
 		//assert(col->size() == Dal::zero );
  		for( unsigned long j = 0; j < col -> rows(); ++j )
 		{
     			//CellData * c = col -> cellData( j );
     			ColumnData * c = col -> data();
     			assert( c -> parent() -> name() == col -> name() );
     			//assert( c -> size() == 0 );
     			//c -> size( 10 + j );
     			//assert( c -> size() == 10 + j );
			const unsigned long n = nRows;
			switch( col -> dataType() )
			{
				case Column::Bool :
     					//for( unsigned long k = 0; k < n; ++k )
    					//{
     					c -> boolData()[ j ] = j % 2 ? false : true;
    					//} // for
					break;
				case Column::Int8 :
					fillData( c -> int8Data(), n );
					break;
				case Column::Int16 :
					fillData( c -> int16Data(), n );
					break;
				case Column::Int32 :
					fillData( c -> int32Data(), n );
					break;
				case Column::Real32 :
					fillData( c -> real32Data(), n );
					break;
				case Column::Real64 :
					fillData( c -> real64Data(), n );
					break;
				case Column::DString :
					//for( unsigned int j = 0; k < n; ++k )
					c -> stringData()[ j ] = 'a' + j % 26;
					break;
				default :
					break;
			} // switch
			delete c;
		} // for
	} // for
}


template<class T>
static void
checkData( T * data, unsigned long size )
{
     for( unsigned long i = 0; i < size; ++i )
     {
     	assert( data[ i ] == T(i + size) );
     } // for
}


static void
check( Table * table )
{
	for( unsigned int i = 0; i < table -> columns(); ++i )
	{
		Column * col = table -> column( i );
  		assert(col->elements() == 1);
 		//assert(col->size() == Dal::zero );
  		for( unsigned long j = 0; j < col -> rows(); ++j )
 		{
     			ColumnData * c = col -> data();
     			assert( c -> parent() -> name() == col -> name() );
     			//assert( c -> size() == 0 );
     			//assert( c -> size() == 10 + j );
			const unsigned long n = nRows;
			switch( col -> dataType() )
			{
				case Column::Bool :
     					//for( unsigned long k = 0; k < n; ++k )
    					//{
						//cout << int((c -> boolData()[ k ])) << " " << int(bool8( k % 2 ? false : true )) << endl;
     						assert( c -> boolData()[ j ] == bool8( j % 2 ? false : true ));
    					//} // for
					break;
				case Column::Int8 :
					checkData( c -> int8Data(), n );
					break;
				case Column::Int16 :
					checkData( c -> int16Data(), n );
					break;
				case Column::Int32 :
					checkData( c -> int32Data(), n );
					break;
				case Column::Real32 :
					checkData( c -> real32Data(), n );
					break;
				case Column::Real64 :
					checkData( c -> real64Data(), n );
					break;
				case Column::DString :
					//for( unsigned int k = 0; k < n; ++k )
					//{
					//	cout << ((c -> stringData()[ k ])) << " " << char('a' + (j + k) % 26 ) << endl;
						assert( c -> stringData()[ j ] == char('a' + j % 26 ));
					//} // for
					break;
				default :
					break;
			} // switch
     			//assert( c -> size() == 10 + j );
			delete c;
		} // for
	} // for
}


static void
basicChecks( Table * table )
{
  assert( table -> rows() == nRows );
  assert( table -> columns() == numberOfColumns );
  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
  	Column* col = table -> column( i );
  	assert( col -> dataType() == cols[ i ].type );
  	assert( col -> name() == cols[ i ].name );
  	assert( col -> units() == cols[ i ].units );
  	assert( col -> label() == cols[ i ].label );
	assert( col -> attributes() == 2 );
	Attribute* att = col -> attribute( 0 );
	assert( att -> units() == "mm" );
	assert( att -> label() == "small attribute" );
	assert( att -> name() == "TLMIN" );
	att = col -> attribute( 1 );
	assert( att -> name() == "TLMAX" );
	assert( att -> units() == "km" );
	assert( att -> label() == "large attribute" );
  } // for
}


static void
test( unsigned long numberOfRows )
{

  nRows = numberOfRows;

  DataSet* set = dataSetServer->open("test.dat",DataSet::Create);
  Table* tab = set->addTable("tab1",nRows);
  for( unsigned int i = 0; i < numberOfColumns; ++i )
  {
  	Column* col = tab->addColumn(
			cols[ i ].name,
			cols[ i ].type,
			cols[ i ].label,
			cols[ i ].units );
	Attribute* att = col -> addAttribute(
					"TLMIN",
					"small attribute", 
					"mm" );
	*att = 2 * int(col -> columnNumber() );
	att = col -> addAttribute(
					"TLMAX",
					"large attribute", 
					"km" );
	*att = int(col -> columnNumber() );
  } // for

  // Basic checks
  basicChecks( tab );
  // Populate the columns
  fill( tab );
  // Checks
  check( tab );

  dataSetServer->close(set);
  set = dataSetServer->open("test.dat",DataSet::Read);
  tab = set -> table( 0 );
  // Basic checks
  basicChecks( tab );
  // Checks
  check( tab );
  dataSetServer->close(set);

  set = dataSetServer->open("test.dat",DataSet::Modify);
  tab = set -> table( 0 );
  // Basic checks
  basicChecks( tab );
  // Checks
  check( tab );
  dataSetServer->close(set);

  // ******************************************************
  // Create a second set, and add the table from the first.
  // ******************************************************
  set = dataSetServer->open("test.dat",DataSet::Modify);
  DataSet* set2 = dataSetServer->open("test2.dat",DataSet::Create);

  tab = set -> table( 0 );
  set->add(tab,"tab2");
  Table* tab2 = set -> table( 1 );
  assert( tab2 -> name() == "tab2" );
  basicChecks( tab2 );
  check( tab2 );

  set2->add(tab,"tab3");
  Table* tab3 = set2 -> table( 0 );
  assert( tab3 -> name() == "tab3" );
  basicChecks( tab3 );
  check( tab3 );

  dataSetServer->close(set);
  dataSetServer->close(set2);
}


int main()
{

  test( 0 );
  test( 1 );
  test( 10 );
  test( 100 );

  return 0;
}
