using namespace std;
#include<cassert>

#include <Dal.h>
#include <sstream>

void
test( unsigned long nRows )
{
  // Create the set
  DataSet* set = dataSetServer->open("test.dat",DataSet::Create);
  Table* tab = set->addTable("some table",nRows);
  Column *col;
  ColumnData *d1;
  ColumnData *d2;
  int32 *i32;

  assert( tab -> columns() == 0 );
  col = tab->addColumn("int32",Column::Int32,"int32 data","dm");
  assert( tab -> columns() == 1 );
  i32 = col -> data() -> int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    d1 = col->data();
    i32 = d1->int32Data();
    i32[i] = 3*i;
    delete d1;
  } // for

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( col->data()->int32Data()[i] == int32( 3*i ));
    assert( col->data()->int32Data()[0] == col->data()->int32Data()[0] );
  } // for

  col = tab -> addColumn("col3",Column::Int32,"int32 data","dm");
  assert( tab -> columns() == 2 );
  d1 = col -> data();
  i32 = d1 -> int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    i32[i] = 3*i;
  } // for

  dataSetServer->close(set);

  // Read the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  tab = set->table("some table");
  assert( tab -> columns() == 2 );
  col = tab->column("int32");

  for (unsigned long i = 0; i < nRows; ++i)
  {
    d1 = col->data();
    i32 = d1->int32Data();
    assert( i32[i] == int32( 3*i ));
    d2 = col->data();
    i32 = d2->int32Data();
    assert( i32[i] == int32( 3*i ));
    delete d1;
    delete d2;
  } // for
   
  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( col->data()->int32Data()[i] == int32( 3*i ));
    assert( col->data()->int32Data()[i] == col->data()->int32Data()[i] );
  } // for
  
  for (unsigned long i = 0; i < nRows; ++i)
  {
    col->data()->int32Data()[i] = 4*i;
    assert( col->data()->int32Data()[i] == int32( 4*i ) );
    assert( col->data()->int32Data()[i] == col->data()->int32Data()[i] );
  } // for
 
  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( col->data(i,1)->int32Data()[0] == int32( 4*i ));
    assert( col->data(i,1)->int32Data()[0] == col->data(i,1)->int32Data()[0] );
  } // for
 
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  tab = set->table("some table");
  assert( tab -> columns() == 2 );
  col = tab->column("int32");

  for (unsigned long i = 0; i < nRows; ++i)
  {
    d1 = col->data();
    i32 = d1->int32Data();
    assert( i32[i] == int32( 4*i ));
    d2 = col->data();
    i32 = d2->int32Data();
    assert( i32[i] == int32( 4*i ));
    delete d1;
    delete d2;
  } // for
   
  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( col->data()->int32Data()[i] == int32( 4*i ));
    assert( col->data()->int32Data()[i] == col->data()->int32Data()[i] );
  } // for
  
  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( col->data(i,1)->int32Data()[0] == int32( 4*i ));
    assert( col->data(i,1)->int32Data()[0] == col->data(i,1)->int32Data()[0] );
  } // for
 
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  i32 =
  	set -> table("some table" ) ->
	column("int32") ->
	data() ->
	int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( i32[i] == int32( 4*i ));
  } // for
   
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  i32 =
  	set -> table("some table" ) ->
	column("int32") ->
	data() ->
	int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    i32[i] = int32( 5*i );
  } // for
   
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  i32 =
  	set -> table("some table" ) ->
	column("int32") ->
	data() ->
	int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( i32[i] == int32( 5*i ));
  } // for
   
  dataSetServer->close(set);

  // Modify the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  tab = set -> table("some table");
  assert( tab -> columns() == 2 );
  col = tab -> addColumn("col2",Column::Int32,"int32 data","dm");
  assert( tab -> columns() == 3 );
  d1 = col -> data();
  i32 = d1 -> int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
  	i32[i] = 3*i;
  } // for

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( i32[i] == int32( 3*i ));
  } // for
   
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Read);
  i32 =
  	set->table("some table" ) ->
	column("col2") ->
	data() ->
	int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( i32[i] == int32( 3*i ));
  } // for
   
  dataSetServer->close(set);

  // Modify the set
  set = dataSetServer->open("test.dat",DataSet::Modify);
  tab = set -> table("some table" );
  assert( tab -> columns() == 3 );
  i32 = tab -> addColumn("col4",Column::Int32,"int32 data","dm") -> data()
	 -> int32Data();
  assert( tab -> columns() == 4 );

  for (unsigned long i = 0; i < nRows; ++i)
  {
  	i32[i] = 3*i;
  } // for

  for (unsigned long i = 0; i < nRows; ++i)
  {
    i32[i] = int32( 6*i );
  } // for
   
  dataSetServer->close(set);

  // Re-Read the set
  set = dataSetServer->open("test.dat",DataSet::Read);
  i32 =
  	set->table("some table" ) ->
	column("col4") ->
	data() ->
	int32Data();

  for (unsigned long i = 0; i < nRows; ++i)
  {
    assert( i32[i] == int32( 6*i ));
  } // for
   
  dataSetServer->close(set);
}

int
main()
{

  test( 1 );
  test( 2 );
  test( 100 );

  return 0;
}
