using namespace std;
#include<cassert>

#include "Tracer.h"
#include "Dal.h"


static void fill( Column* col ){
    for( unsigned int i = 0; i < col -> rows(); ++i ){
	ColumnData* cd = col -> rowData( i, Dal::AsParent );
	int32 *i32 = cd -> int32Data();
	for( unsigned long j = 0; j < cd -> elements(); ++j ){
	    i32[ j ] = 3 * i + j;
	}
	delete cd;
    } 
}

static void check( Column* col )
{
	PROC
	for( unsigned int i = 0; i < col -> rows(); ++i )
	{
  		ColumnData* cd = col -> rowData( i, Dal::AsParent );
	  	int32 *i32 = cd -> int32Data();
// 		std::cerr << " row " << i << " elements "
// 			  << cd->elements() << std::endl;
  		for( unsigned long j = 0; j < cd -> elements(); ++j )
		{
// 		    std::cerr << " row " << i << " index " << j
// 			      << " value " << i32[j] << " expect "
// 			      << 3*i + j << std::endl;
    			assert( i32[ j ] == int32( 3 * i + j ) );
		}
  		delete cd;
	} // if
	ENDPROC
}


int
main()
{
	PROC


  	DataSet* set = dataSetServer->open("test.dat", DataSet::Create);
  	Table* table = set->addTable("FIRST", 100 );
	Column * col = 0;

  	col = table->addColumn("VAR", 
				  Column::Int32, 
				  "No comment", 
				  "", 
				  Dal::zero);

	for( unsigned int i = 0; i < col -> rows(); ++i )
	{
		CellData* cd = col -> cellData( i );
		cd -> size( i + 1 );
		delete cd;
	} // for

	std::cerr << "Filling variable length column\n";
	fill( col );

	std::cerr << "Checking variable length column\n";
	check( col );

	col = table -> addColumn( "FIX", Column::Int32,"int32 data","dm" );
	std::cerr << "Filling fixed length column\n";
	fill( col );
	std::cerr << "Checking fixed length column\n";
	check( col );
	dataSetServer -> close( set );
	std::cerr << "Dataset closed\n";

  	set = dataSetServer->open("test.dat", DataSet::Read);
	std::cerr << "Dataset opened\n";
  	table = set->table("FIRST");
	col = table -> column("VAR");
	std::cerr << "Checking variable length column\n";
	check(col);
	col = table -> column( "FIX" );
	std::cerr << "Checking fixed length column\n";
	check( col );
	dataSetServer -> close( set );

	ENDPROC
}
