program test

  use dal
  use errorhandling

  implicit none

  type(DataSetT) set
  type(ArrayT) arr
  integer(kind=int32), dimension(:,:,:), pointer :: a
  integer(kind=int32), dimension(:), pointer :: b
  integer, dimension(3), parameter :: s = (/ 3,4,2 /)
  integer :: i,j,k,n

  ! create a set
  set = dataSet("test.dat",CREATE)
  arr = addArray(set, "some table", INTEGER32, dimensions=s )
  call setAttribute(arr,"RA","(none)","RA of plate centre")

  a => int32array3Data(arr)

  ! test whether the shape of the returned array is identical to the shape
  ! originally specified
  if (any(shape(a) /= s)) then
    write(errstr,*) "shape of returned array",shape(a),"/= array spec",s
    call error("internalError")
  endif

  if (any(dimensions(arr) /= s)) then
    write(errstr,*) "dimensions of array",dimensions(arr),"/= array spec",s
    call error("internalError")
  endif

  ! fill with unique numbers
  n = 0
  do k=0,1
    do j=0,3
      do i=0,2 
        a(i,j,k) = n
	n = n + 1
      end do
    end do
  end do

  write(*,*) a

  call release(arr)

  ! access it again as vector
  arr = array(set,"some table",READ);
  b => int32data(arr)
  write(*,*) b

  ! check the shape
  if ( any(shape(b) /= (/24/)) ) then
    write(errstr,*) "vector size",shape(b),"unequal to array size",s
    call error("internalError")
  endif
  
  ! check the values
  do i=0,23
    if ( b(i) /= i ) then
      write(errstr,*) "vector value",b(n),"unequal to array value",n
      call error("internalError")
    end if
  end do

  call release(arr)

  call release(set)

end program test
