program test

  use dal
  use errorhandling

  implicit none

  integer, parameter :: nRows = 1000
  integer, parameter :: nElements = 5
  integer, dimension(1) :: size1 = (/ 5 /)
  integer, dimension(2) :: size2 = (/ 5,2 /)
  real(kind=single), dimension(:,:), pointer :: r32
  integer(kind=int32), dimension(:,:,:), pointer :: i32
  integer(kind=int32) :: ival
  real(kind=single) :: rval
  integer :: i, j, k

  type(DataSetT) :: set
  type(TableT) :: tab
  type(ColumnT) :: col

  set = dataSet("test.dat",Read)
  tab = table(set,"someTable")

  if (numberOfRows(tab) /= nRows) call error("test","nRows")
  
  col = column(tab,"real32",Read)
  write(*,*) dimensions(col)
  if (any(dimensions(col) /= size1)) call error("test","real dimensions")

  r32 => real32VectorData(col)
  write(*,*) shape(r32)
  if (size(r32,2) /= numberOfRows(tab)) call error("test","shape r32 2")
  if (size(r32,1) /= size1(1)) call error("test","shape r32 1")

  do i=0,size(r32,1)-1
    do j=0,size(r32,2)-1
      rval = i + 0.1 * j
      if (abs(r32(i,j)-rval) > 1e-5) call error("test","r32 value")
    end do
  end do

  col = column(tab,"int32",Read);
  write(*,*) dimensions(col)
  !if (any(dimensions(col) /= size2)) call error("test","int dimensions")

  i32 => int32Array2Data(col);
  write(*,*) shape(i32)
  if (size(i32,3) /= numberOfRows(tab)) call error("test","shape i32 3")
  if (size(i32,2) /= size2(2)) call error("test","shape i32 2")
  if (size(i32,1) /= size2(1)) call error("test","shape i32 1")

  do i=0,size(i32,1)-1
    do j=0,size(i32,2)-1
      do k=0,size(i32,3)-1
        ival = 100*i + 10*j + k
	if (i32(i,j,k) /= ival) call error("test","i32 value")
      end do
    end do
  end do

  call release(set)

end program
