using namespace std;
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine science & computing) 
// September 2001
//

// 2001-09-29: this is a cifitsio bug. Waiting for a fix from Bill Pence.

#include<Dal.h>
#include<iostream>

#define LONG_CODE 0

#if LONG_CODE
#include<errstr.h>
#include <ExceptionErrorHandler.h>  
#include<cstdlib>
#include<cstring>
#include<fitsio.h>
#include<longnam.h>
#endif

int main()
{

#if LONG_CODE
  ExceptionErrorHandler local;
  ErrorHandler* oldHandler = errHandler.handler(&local);
#endif

  // SOC-SPR-2149
  // Create a table with a variable length column
  // Create another table
  // Try to access and write the data in the first table
  // -> crash in low mem
  // Moving the addition of the second table to after the access of
  // data from the first table solves the problem.
#if LONG_CODE
  try
    {
#endif
      std::cout << "DAL running" << std::endl;
      const size_t breakpoint = 719;  
      struct DataSet *s = dataSetServer->open("scratch.ds",Dal::Create);
      
      struct Table *t = s->addTable("FIRST",1);
      
      s->addTable("SECOND",1);
      
      struct Column *c = t->addColumn("TEST",Dal::Real32,"","",Dal::zero);
      struct CellData *d = c->cellData(0);
      d->size(breakpoint);
      delete d;
      
      dataSetServer->close(s);
#if LONG_CODE
    }
  catch(...){
    std::cout << "The DAL failed" << std::endl;
  }

  // Try to do the same with cfitsio itself
  {
    std::cout << "cfitsio running" << std::endl;

    int N = 1000;
    int *arr, i;
    int status = 0;
    fitsfile *fptr;
    char *ttype1[] = { "Ivalue", };
    char *tform1[] = { "1PJ", };
    char *tunit1[] = { "s", };
    char name1[] = "X1";
    arr = static_cast<int *>(malloc(N*sizeof(int)));
    for (i=0; i<N; ++i)
      arr[i] = i;
    fits_create_file(&fptr, "!cfits.fits", &status);
    printf("%d", status);
    fits_create_tbl(fptr, BINARY_TBL, 0, 1, ttype1, tform1, tunit1,
		    name1, &status);

    char *ttype2[] = { "Ivalue", };
    char *tform2[] = { "1J", };
    char *tunit2[] = { "s", };
    char name2[] = "X2";
    fits_create_tbl(fptr, BINARY_TBL, 0, 1, ttype2, tform2, tunit2,
		    name2, &status);

    fits_write_col(fptr, TINT, 1, 1, 1, N, arr, &status);
    fits_movnam_hdu(fptr, BINARY_TBL, name1, 0, &status);
    fits_write_col(fptr, TINT, 1, 1, 1, N, arr, &status);
    fits_close_file(fptr, &status);        
    printf("%d", status);
  }
#endif
}

  
